from sympy import pi

from pyeuclid.formalization.relation import *
from pyeuclid.formalization.utils import *

construction_rule_sets = {}


class ConstructionRule:
    """Base class for geometric construction rules."""

    def __init__(self):
        """Initialize an empty construction rule."""
        pass
    
    def arguments(self):
        """Return the input entities required for the construction."""
        return []
        
    def constructed_points(self):
        """Return the points constructed by this rule."""
        return []

    def conditions(self):
        """Return prerequisite relations for the construction to be valid."""
        return []
    
    def conclusions(self):
        """Return relations implied after applying the construction."""
        return []

    def __str__(self):
        class_name = self.__class__.__name__
        attributes = ",".join(str(value) for _, value in vars(self).items())
        return f"{class_name}({attributes})"


class register:
    """Decorator that registers a construction rule into labeled sets."""

    def __init__(self, *annotations):
        self.annotations = annotations

    def __call__(self, cls):
        for item in self.annotations:
            if not item in construction_rule_sets:
                construction_rule_sets[item] = [cls]
            else:
                construction_rule_sets[item].append(cls)

        def expanded_conditions(self):
            return expand_definition(self._conditions())

        cls._conditions = cls.conditions
        cls.conditions = expanded_conditions
        return cls


@register("AG")
class construct_angle_bisector(ConstructionRule):
    """Construct the bisector point X of angle ABC."""
    def __init__(self, x, a, b, c):
        self.x, self.a, self.b, self.c = x, a, b, c
        
    def arguments(self):
        return [self.a, self.b, self.c]
    
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [NotCollinear(self.a, self.b, self.c)]

    def conclusions(self):
        return [Angle(self.a, self.b, self.x) - Angle(self.x, self.b, self.c)]


@register("AG")
class construct_angle_mirror(ConstructionRule):
    """Construct point X as the mirror of BA across BC."""
    def __init__(self, x, a, b, c):
        self.x, self.a, self.b, self.c = x, a, b, c
        
    def arguments(self):
        return [self.a, self.b, self.c]
    
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [NotCollinear(self.a, self.b, self.c)]

    def conclusions(self):
        return [Angle(self.a, self.b, self.c) - Angle(self.c, self.b, self.x)]


@register("AG")
class construct_circle(ConstructionRule):
    """Construct circle center X equidistant from A, B, C."""
    def __init__(self, x, a, b, c):
        self.x, self.a, self.b, self.c = x, a, b, c
        
    def arguments(self):
        return [self.a, self.b, self.c]
    
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [NotCollinear(self.a, self.b, self.c)]

    def conclusions(self):
        return [
            Length(self.x, self.a) - Length(self.x, self.b),
            Length(self.x, self.b) - Length(self.x, self.c),
        ]


@register("AG")
class construct_circumcenter(ConstructionRule):
    """Construct circumcenter X of triangle ABC."""
    def __init__(self, x, a, b, c):
        self.x, self.a, self.b, self.c = x, a, b, c
        
    def arguments(self):
        return [self.a, self.b, self.c]
    
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [NotCollinear(self.a, self.b, self.c)]

    def conclusions(self):
        return [
            Length(self.x, self.a) - Length(self.x, self.b),
            Length(self.x, self.b) - Length(self.x, self.c),
        ]


@register("AG")
class construct_eq_quadrangle(ConstructionRule):
    """Construct quadrilateral ABCD with equal diagonals."""
    def __init__(self, a, b, c, d):
        self.a, self.b, self.c, self.d = a, b, c, d
        
    def constructed_points(self):
        return [self.a, self.b, self.c, self.d]
    
    def conclusions(self):
        return [
            Length(self.a, self.d) - Length(self.b, self.c)
        ]


@register("AG")
class construct_eq_trapezoid(ConstructionRule):
    """Construct isosceles trapezoid ABCD (AB ∥ CD)."""
    def __init__(self, a, b, c, d):
        self.a, self.b, self.c, self.d = a, b, c, d
        
    def constructed_points(self):
        return [self.a, self.b, self.c, self.d]

    def conclusions(self):
        return [
            Length(self.a, self.d) - Length(self.b, self.c),
            Parallel(self.a, self.b, self.c, self.d),
            Angle(self.d, self.a, self.b) - Angle(self.a, self.b, self.c),
            Angle(self.b, self.c, self.d) - Angle(self.c, self.d, self.a),
        ]


@register("AG")
class construct_eq_triangle(ConstructionRule):
    """Construct equilateral triangle with vertex X and base BC."""
    def __init__(self, x, b, c):
        self.x, self.b, self.c = x, b, c
        
    def arguments(self):
        return [self.b, self.c]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [Different(self.b, self.c)]

    def conclusions(self):
        return [
            Length(self.x, self.b) - Length(self.b, self.c),
            Length(self.b, self.c) - Length(self.c, self.x),
            Angle(self.x, self.b, self.c) - Angle(self.b, self.c, self.x),
            Angle(self.c, self.x, self.b) - Angle(self.x, self.b, self.c),
        ]


@register("AG")
class construct_eqangle2(ConstructionRule):
    """Construct X so that angle ABX equals angle XCB."""
    def __init__(self, x, a, b, c):
        self.x, self.a, self.b, self.c = x, a, b, c
        
    def arguments(self):
        return [self.a, self.b, self.c]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [NotCollinear(self.a, self.b, self.c)]

    def conclusions(self):
        return [Angle(self.b, self.a, self.x) - Angle(self.x, self.c, self.b)]


@register("AG")
class construct_eqdia_quadrangle(ConstructionRule):
    def __init__(self, a, b, c, d):
        self.a, self.b, self.c, self.d = a, b, c, d
        
    def constructed_points(self):
        return [self.a, self.b, self.c, self.d]
    
    def conclusions(self):
        return [
            Length(self.b, self.d) - Length(self.a, self.c)
        ]


@register("AG")
class construct_eqdistance(ConstructionRule):
    def __init__(self, x, a, b, c):
        self.x, self.a, self.b, self.c = x, a, b, c
    
    def arguments(self):
        return [self.a, self.b, self.c]
        
    def constructed_points(self):
        return [self.x]
    
    def conditions(self):
        return [Different(self.b, self.c)]

    def conclusions(self):
        return [Length(self.x, self.a) - Length(self.b, self.c)]


@register("AG")
class construct_eqdistance2(ConstructionRule):
    def __init__(self, x, a, b, c, alpha):
        self.x, self.a, self.b, self.c, self.alpha = x, a, b, c, alpha
    
    def arguments(self):
        return [self.a, self.b, self.c, self.alpha]
        
    def constructed_points(self):
        return [self.x]
    
    def conditions(self):
        return [Different(self.b, self.c)]

    def conclusions(self):
        return [Length(self.x, self.a) - sympy.simplify(self.alpha) * Length(self.b, self.c)]


@register("AG")
class construct_eqdistance3(ConstructionRule):
    def __init__(self, x, a, alpha):
        self.x, self.a, self.alpha = x, a, alpha
    
    def arguments(self):
        return [self.a, self.alpha]
        
    def constructed_points(self):
        return [self.x]

    def conclusions(self):
        return [Length(self.x, self.a) - sympy.simplify(self.alpha)]


@register("AG")
class construct_foot(ConstructionRule):
    def __init__(self, x, a, b, c):
        self.x, self.a, self.b, self.c = x, a, b, c
    
    def arguments(self):
        return [self.a, self.b, self.c]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [NotCollinear(self.a, self.b, self.c)]

    def conclusions(self):
        return [Perpendicular(self.x, self.a, self.b, self.c), Collinear(self.x, self.b, self.c)]


@register("AG")
class construct_free(ConstructionRule):
    def __init__(self, a):
        self.a = a
        
    def constructed_points(self):
        return [self.a]


@register("AG")
class construct_incenter(ConstructionRule):
    def __init__(self, x, a, b, c):
        self.x, self.a, self.b, self.c = x, a, b, c
    
    def arguments(self):
        return [self.a, self.b, self.c]
        
    def constructed_points(self):
        return [self.x]
    
    def conditions(self):
        return [NotCollinear(self.a, self.b, self.c)]

    def conclusions(self):
        return [
            Angle(self.b, self.a, self.x) - Angle(self.x, self.a, self.c),
            Angle(self.a, self.c, self.x) - Angle(self.x, self.c, self.b),
            Angle(self.c, self.b, self.x) - Angle(self.x, self.b, self.a),
        ]


class construct_incenter2(ConstructionRule):
    def __init__(self, x, y, z, i, a, b, c):
        self.x, self.y, self.z, self.i, self.a, self.b, self.c = x, y, z, i, a, b, c
        
    def arguments(self):
        return [self.a, self.b, self.c]
        
    def constructed_points(self):
        return [self.x, self.y, self.z, self.i]
    
    def conditions(self):
        return [NotCollinear(self.a, self.b, self.c)]

    def conclusions(self):
        return [
            Angle(self.b, self.a, self.i) - Angle(self.i, self.a, self.c),
            Angle(self.a, self.c, self.i) - Angle(self.i, self.c, self.b),
            Angle(self.c, self.b, self.i) - Angle(self.i, self.b, self.a),
            Collinear(self.x, self.b, self.c),
            Perpendicular(self.i, self.x, self.b, self.c),
            Collinear(self.y, self.c, self.a),
            Perpendicular(self.i, self.y, self.c, self.a),
            Collinear(self.z, self.a, self.b),
            Perpendicular(self.i, self.z, self.a, self.b),
            Length(self.i, self.x) - Length(self.i, self.y),
            Length(self.i, self.y) - Length(self.i, self.z),
        ]


@register("AG")
class construct_excenter(ConstructionRule):
    def __init__(self, x, a, b, c):
        self.x, self.a, self.b, self.c = x, a, b, c
    
    def arguments(self):
        return [self.a, self.b, self.c]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [NotCollinear(self.a, self.b, self.c)]

    def conclusions(self):
        return [
            Angle(self.b, self.a, self.x) - Angle(self.x, self.a, self.c),
            Angle(self.a, self.b, self.x) - (Angle(self.a, self.b, self.c) / 2 + pi / 2),
            Angle(self.a, self.c, self.x) - (Angle(self.a, self.c, self.b) / 2 + pi / 2),
        ]


class construct_excenter2(ConstructionRule):
    def __init__(self, x, y, z, i, a, b, c):
        self.x, self.y, self.z, self.i, self.a, self.b, self.c = x, y, z, i, a, b, c
    
    def arguments(self):
        return [self.a, self.b, self.c]
        
    def constructed_points(self):
        return [self.x, self.y, self.z, self.i]
    
    def conditions(self):
        return [NotCollinear(self.a, self.b, self.c)]

    def conclusions(self):
        return [
            Angle(self.b, self.a, self.i) - Angle(self.i, self.a, self.c),
            Angle(self.a, self.b, self.i) - (Angle(self.a, self.b, self.c) / 2 + pi / 2),
            Angle(self.a, self.c, self.i) - (Angle(self.a, self.c, self.b) / 2 + pi / 2),
            Collinear(self.x, self.b, self.c),
            Perpendicular(self.i, self.x, self.b, self.c),
            Collinear(self.y, self.c, self.a),
            Perpendicular(self.i, self.y, self.c, self.a),
            Collinear(self.z, self.a, self.b),
            Perpendicular(self.i, self.z, self.a, self.b),
            Length(self.i, self.x) - Length(self.i, self.y),
            Length(self.i, self.y) - Length(self.i, self.z),
        ]


@register("AG")
class construct_centroid(ConstructionRule):
    def __init__(self, x, y, z, i, a, b, c):
        self.x, self.y, self.z, self.i, self.a, self.b, self.c = x, y, z, i, a, b, c
    
    def arguments(self):
        return [self.a, self.b, self.c]
        
    def constructed_points(self):
        return [self.x, self.y, self.z, self.i]
    
    def conditions(self):
        return [NotCollinear(self.a, self.b, self.c)]

    def conclusions(self):
        return [
            Collinear(self.x, self.b, self.c),
            Length(self.x, self.b) - Length(self.x, self.c),
            Collinear(self.y, self.c, self.a),
            Length(self.y, self.c) - Length(self.y, self.a),
            Collinear(self.z, self.a, self.b),
            Length(self.z, self.a) - Length(self.z, self.b),
            Collinear(self.a, self.x, self.i),
            Collinear(self.b, self.y, self.i),
            Collinear(self.c, self.z, self.i),
        ]


@register("AG")
class construct_intersection_cc(ConstructionRule):
    def __init__(self, x, o, w, a):
        self.x, self.o, self.w, self.a = x, o, w, a
    
    def arguments(self):
        return [self.o, self.w, self.a]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [NotCollinear(self.o, self.w, self.a)]

    def conclusions(self):
        return [
            Length(self.o, self.a) - Length(self.o, self.x),
            Length(self.w, self.a) - Length(self.w, self.x),
        ]


@register("AG")
class construct_intersection_lc(ConstructionRule):
    def __init__(self, x, a, o, b):
        self.x, self.a, self.o, self.b = x, a, o, b
    
    def arguments(self):
        return [self.a, self.o, self.b]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [
            Different(self.a, self.b),
            Different(self.o, self.b),
            Not(Perpendicular(self.b, self.o, self.b, self.a)),
        ]

    def conclusions(self):
        return [
            Collinear(self.x, self.a, self.b),
            Length(self.o, self.b) - Length(self.o, self.x),
        ]


@register("AG")
class construct_intersection_ll(ConstructionRule):
    def __init__(self, x, a, b, c, d):
        self.x, self.a, self.b, self.c, self.d = x, a, b, c, d
    
    def arguments(self):
        return [self.a, self.b, self.c, self.d]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [
            Not(Parallel(self.a, self.b, self.c, self.d))
            # ,Not(Collinear(self.a,self.b,self.c,self.d)) # TODO
        ]

    def conclusions(self):
        return [Collinear(self.x, self.a, self.b), Collinear(self.x, self.c, self.d)]


@register("AG")
class construct_intersection_lp(ConstructionRule):
    def __init__(self, x, a, b, c, m, n):
        self.x, self.a, self.b, self.c, self.m, self.n = x, a, b, c, m, n
    
    def arguments(self):
        return [self.a, self.b, self.c, self.m, self.n]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [
            Not(Parallel(self.m, self.n, self.a, self.b)),
            NotCollinear(self.a, self.b, self.c),
            NotCollinear(self.c, self.m, self.n),
        ]

    def conclusions(self):
        return [Collinear(self.x, self.a, self.b), Parallel(self.c, self.x, self.m, self.n)]


@register("AG")
class construct_intersection_lt(ConstructionRule):
    def __init__(self, x, a, b, c, d, e):
        self.x, self.a, self.b, self.c, self.d, self.e = x, a, b, c, d, e
        
    def arguments(self):
        return [self.a, self.b, self.c, self.d, self.e]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [
            NotCollinear(self.a, self.b, self.c),
            Not(Perpendicular(self.a, self.b, self.d, self.e)),
        ]

    def conclusions(self):
        return [Collinear(self.x, self.a, self.b), Perpendicular(self.x, self.c, self.d, self.e)]


@register("AG")
class construct_intersection_pp(ConstructionRule):
    def __init__(self, x, a, b, c, d, e, f):
        self.x, self.a, self.b, self.c, self.d, self.e, self.f = x, a, b, c, d, e, f
    
    def arguments(self):
        return [self.a, self.b, self.c, self.d, self.e, self.f]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [
            Different(self.a, self.d),
            Not(Parallel(self.b, self.c, self.e, self.f)),
        ]

    def conclusions(self):
        return [
            Parallel(self.x, self.a, self.b, self.c),
            Parallel(self.x, self.d, self.e, self.f),
        ]


@register("AG")
class construct_intersection_tt(ConstructionRule):
    def __init__(self, x, a, b, c, d, e, f):
        self.x, self.a, self.b, self.c, self.d, self.e, self.f = x, a, b, c, d, e, f
    
    def arguments(self):
        return [self.a, self.b, self.c, self.d, self.e, self.f]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [
            Different(self.a, self.d),
            Not(Parallel(self.b, self.c, self.e, self.f)),
        ]

    def conclusions(self):
        return [
            Perpendicular(self.x, self.a, self.b, self.c),
            Perpendicular(self.x, self.d, self.e, self.f),
        ]


@register("AG")
class construct_iso_triangle(ConstructionRule):
    def __init__(self, a, b, c):
        self.a, self.b, self.c = a, b, c
        
    def constructed_points(self):
        return [self.a, self.b, self.c]
    
    def conclusions(self):
        return [Length(self.a, self.b) - Length(self.a, self.c), Angle(self.a, self.b, self.c) - Angle(self.b, self.c, self.a)]


@register("AG")
class construct_lc_tangent(ConstructionRule):
    def __init__(self, x, a, o):
        self.x, self.a, self.o = x, a, o
    
    def arguments(self):
        return [self.a, self.o]
        
    def constructed_points(self):
        return [self.x]

    def conclusions(self):
        return [Perpendicular(self.a, self.x, self.a, self.o)]


@register("AG")
class construct_midpoint(ConstructionRule):
    def __init__(self, x, a, b):
        self.x, self.a, self.b = x, a, b
    
    def arguments(self):
        return [self.a, self.b]
        
    def constructed_points(self):
        return [self.x]
    
    def conditions(self):
        return [Different(self.a, self.b)]

    def conclusions(self):
        return [Collinear(self.x, self.a, self.b), Length(self.x, self.a) - Length(self.x, self.b)]


@register("AG")
class construct_mirror(ConstructionRule):
    def __init__(self, x, a, b):
        self.x, self.a, self.b = x, a, b
    
    def arguments(self):
        return [self.a, self.b]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [Different(self.a, self.b)]

    def conclusions(self):
        return [
            Length(self.b, self.a) - Length(self.b, self.x),
            Collinear(self.x, self.a, self.b),
        ]


@register("AG")
class construct_nsquare(ConstructionRule):
    def __init__(self, x, a, b):
        self.x, self.a, self.b = x, a, b
    
    def arguments(self):
        return [self.a, self.b]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [Different(self.a, self.b)]

    def conclusions(self):
        return [
            Length(self.x, self.a) - Length(self.a, self.b),
            Perpendicular(self.x, self.a, self.a, self.b),
        ]


@register("AG")
class construct_on_aline(ConstructionRule):
    def __init__(self, x, a, b, c, d, e):
        self.x, self.a, self.b, self.c, self.d, self.e = x, a, b, c, d, e
    
    def arguments(self):
        return [self.a, self.b, self.c, self.d, self.e]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [
            NotCollinear(self.c, self.d, self.e),
            Different(self.a, self.b),
            Different(self.c, self.d),
            Different(self.c, self.e),
        ]

    def conclusions(self):
        return Angle(self.x, self.a, self.b) - Angle(self.c, self.d, self.e), Angle(self.x, self.a, self.b) + Angle(self.c, self.d, self.e) - pi



# @register("AG")
# class construct_on_aline2(ConstructionRule):
#     def __init__(self, x, a, b, c, d, e):
#         self.x, self.a, self.b, self.c, self.d, self.e = x, a, b, c, d, e
    
#     def arguments(self):
#         return [self.a, self.b, self.c, self.d, self.e]
        
#     def constructed_points(self):
#         return [self.x]

#     def conditions(self):
#         return [
#             NotCollinear(self.c, self.d, self.e),
#             Different(self.a, self.b),
#             Different(self.c, self.d),
#             Different(self.c, self.e),
#         ]

#     def conclusions(self):
#         return [Angle(self.x, self.a, self.b) + Angle(self.c, self.d, self.e) - pi]


@register("AG")
class construct_on_bline(ConstructionRule):
    def __init__(self, x, a, b):
        self.x, self.a, self.b = x, a, b
    
    def arguments(self):
        return [self.a, self.b]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [Different(self.a, self.b)]

    def conclusions(self):
        return [
            Length(self.x, self.a) - Length(self.x, self.b),
            Angle(self.x, self.a, self.b) - Angle(self.a, self.b, self.x),
        ]


@register("AG")
class construct_on_circle(ConstructionRule):
    def __init__(self, x, o, a):
        self.x, self.o, self.a = x, o, a
    
    def arguments(self):
        return [self.o, self.a]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [Different(self.o, self.a)]

    def conclusions(self):
        return [Length(self.o, self.x) - Length(self.o, self.a)]


@register("AG")
class construct_on_line(ConstructionRule):
    def __init__(self, x, a, b):
        self.x, self.a, self.b = x, a, b
    
    def arguments(self):
        return [self.a, self.b]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [Different(self.a, self.b)]

    def conclusions(self):
        return [Collinear(self.x, self.a, self.b)]


@register("AG")
class construct_on_pline(ConstructionRule):
    def __init__(self, x, a, b, c):
        self.x, self.a, self.b, self.c = x, a, b, c
    
    def arguments(self):
        return [self.a, self.b, self.c]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [Different(self.b, self.c), NotCollinear(self.a, self.b, self.c)]

    def conclusions(self):
        return [Parallel(self.x, self.a, self.b, self.c)]


@register("AG")
class construct_on_tline(ConstructionRule):
    def __init__(self, x, a, b, c):
        self.x, self.a, self.b, self.c = x, a, b, c
    
    def arguments(self):
        return [self.a, self.b, self.c]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [Different(self.b, self.c)]

    def conclusions(self):
        return [Perpendicular(self.x, self.a, self.b, self.c)]


@register("AG")
class construct_orthocenter(ConstructionRule):
    def __init__(self, x, a, b, c):
        self.x, self.a, self.b, self.c = x, a, b, c
    
    def arguments(self):
        return [self.a, self.b, self.c]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [NotCollinear(self.a, self.b, self.c)]

    def conclusions(self):
        return [
            Perpendicular(self.x, self.a, self.b, self.c),
            Perpendicular(self.x, self.b, self.c, self.a),
            Perpendicular(self.x, self.c, self.a, self.b),
        ]


@register("AG")
class construct_parallelogram(ConstructionRule):
    def __init__(self, a, b, c, x):
        self.a, self.b, self.c, self.x = a, b, c, x
    
    def arguments(self):
        return [self.a, self.b, self.c]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [NotCollinear(self.a, self.b, self.c)]

    def conclusions(self):
        return [
            Parallel(self.a, self.b, self.c, self.x),
            Parallel(self.a, self.x, self.b, self.c),
            Length(self.a, self.b) - Length(self.c, self.x),
            Length(self.a, self.x) - Length(self.b, self.c),
        ]


@register("AG")
class construct_pentagon(ConstructionRule):
    def __init__(self, a, b, c, d, e):
        self.a, self.b, self.c, self.d, self.e = a, b, c, d, e
        
    def constructed_points(self):
        return [self.a, self.b, self.c, self.d, self.e]


@register("AG")
class construct_psquare(ConstructionRule):
    def __init__(self, x, a, b):
        self.x, self.a, self.b = x, a, b
    
    def arguments(self):
        return [self.a, self.b]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [Different(self.a, self.b)]

    def conclusions(self):
        return [
            Length(self.x, self.a) - Length(self.a, self.b),
            Perpendicular(self.x, self.a, self.a, self.b),
        ]


@register("AG")
class construct_quadrangle(ConstructionRule):
    def __init__(self, a, b, c, d):
        self.a, self.b, self.c, self.d = a, b, c, d
        
    def constructed_points(self):
        return [self.a, self.b, self.c, self.d]


@register("AG")
class construct_r_trapezoid(ConstructionRule):
    def __init__(self, a, b, c, d):
        self.a, self.b, self.c, self.d = a, b, c, d
        
    def constructed_points(self):
        return [self.a, self.b, self.c, self.d]
    
    def conclusions(self):
        return [Perpendicular(self.a, self.b, self.a, self.d), Parallel(self.a, self.b, self.c, self.d)]


@register("AG")
class construct_r_triangle(ConstructionRule):
    def __init__(self, a, b, c):
        self.a, self.b, self.c = a, b, c
        
    def constructed_points(self):
        return [self.a, self.b, self.c]
    
    def conclusions(self):
        return [Perpendicular(self.a, self.b, self.a, self.c)]


@register("AG")
class construct_rectangle(ConstructionRule):
    def __init__(self, a, b, c, d):
        self.a, self.b, self.c, self.d = a, b, c, d
        
    def constructed_points(self):
        return [self.a, self.b, self.c, self.d]
    
    def conclusions(self):
        return [
            Perpendicular(self.a, self.b, self.b, self.c),
            Parallel(self.a, self.b, self.c, self.d),
            Parallel(self.a, self.d, self.b, self.c),
            Perpendicular(self.a, self.b, self.a, self.d),
            Length(self.a, self.b) - Length(self.c, self.d),
            Length(self.a, self.d) - Length(self.b, self.c),
            Length(self.a, self.c) - Length(self.b, self.d),
        ]


@register("AG")
class construct_reflect(ConstructionRule):
    def __init__(self, x, a, b, c):
        self.x, self.a, self.b, self.c = x, a, b, c
    
    def arguments(self):
        return [self.a, self.b, self.c]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [Different(self.b, self.c), NotCollinear(self.a, self.b, self.c)]

    def conclusions(self):
        return [
            Perpendicular(self.b, self.c, self.a, self.x),
            Length(self.a, self.b) - Length(self.b, self.x),
            Length(self.a, self.c) - Length(self.c, self.x),
        ]


@register("AG")
class construct_risos(ConstructionRule):
    def __init__(self, a, b, c):
        self.a, self.b, self.c = a, b, c
        
    def constructed_points(self):
        return [self.a, self.b, self.c]
    
    def conclusions(self):
        return [
            Angle(self.a, self.b, self.c) - Angle(self.b, self.c, self.a),
            Perpendicular(self.a, self.b, self.a, self.c),
            Length(self.a, self.b) - Length(self.a, self.c),
        ]


@register("AG")
class construct_s_angle(ConstructionRule):
    def __init__(self, a, b, x, alpha):
        self.a, self.b, self.x, self.alpha = a, b, x, alpha
    
    def arguments(self):
        return [self.a, self.b, self.alpha]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [Different(self.a, self.b)]

    def conclusions(self):
        return [Angle(self.a, self.b, self.x) - sympy.simplify(sympy.Rational(abs(self.alpha),180)*pi)]


@register("AG")
class construct_segment(ConstructionRule):
    def __init__(self, a, b):
        self.a, self.b = a, b
        
    def constructed_points(self):
        return [self.a, self.b]


@register("AG")
class construct_s_segment(ConstructionRule):
    def __init__(self, a, b, alpha):
        self.a, self.b, self.alpha = a, b, alpha
    
    def arguments(self):
        return [self.alpha]
        
    def constructed_points(self):
        return [self.a, self.b]

    def conclusions(self):
        return [Length(self.a, self.x) - sympy.simplify(self.alpha)]


@register("AG")
class construct_shift(ConstructionRule):
    def __init__(self, x, b, c, d):
        self.x, self.b, self.c, self.d = x, b, c, d
    
    def arguments(self):
        return [self.b, self.c, self.d]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [Different(self.d, self.b)]

    def conclusions(self):
        return [
            Length(self.x, self.b) - Length(self.c, self.d),
            Length(self.x, self.c) - Length(self.b, self.d),
        ]


class construct_square(ConstructionRule):
    def __init__(self, a, b, x, y):
        self.a, self.b, self.x, self.y = a, b, x, y
    
    def arguments(self):
        return [self.a, self.b]
        
    def constructed_points(self):
        return [self.x, self.y]
    
    def conditions(self):
        return [Different(self.a, self.b)]

    def conclusions(self):
        return [
            Perpendicular(self.a, self.b, self.b, self.x),
            Length(self.a, self.b) - Length(self.b, self.x),
            Parallel(self.a, self.b, self.x, self.y),
            Parallel(self.a, self.y, self.b, self.x),
            Perpendicular(self.a, self.y, self.y, self.x),
            Length(self.b, self.x) - Length(self.x, self.y),
            Length(self.x, self.y) - Length(self.y, self.a),
            Perpendicular(self.a, self.x, self.b, self.y),
            Length(self.a, self.x) - Length(self.b, self.y),
        ]


@register("AG")
class construct_isquare(ConstructionRule):
    def __init__(self, a, b, c, d):
        self.a, self.b, self.c, self.d = a, b, c, d
        
    def constructed_points(self):
        return [self.a, self.b, self.c, self.d]
    
    def conclusions(self):
        return [
            Perpendicular(self.a, self.b, self.b, self.c),
            Length(self.a, self.b) - Length(self.b, self.c),
            Parallel(self.a, self.b, self.c, self.d),
            Parallel(self.a, self.d, self.b, self.c),
            Perpendicular(self.a, self.d, self.d, self.c),
            Length(self.b, self.c) - Length(self.c, self.d),
            Length(self.c, self.d) - Length(self.d, self.a),
            Perpendicular(self.a, self.c, self.b, self.d),
            Length(self.a, self.c) - Length(self.b, self.d),
        ]


@register("AG")
class construct_trapezoid(ConstructionRule):
    def __init__(self, a, b, c, d):
        self.a, self.b, self.c, self.d = a, b, c, d
        
    def constructed_points(self):
        return [self.a, self.b, self.c, self.d]
    
    def conclusions(self):
        return [Parallel(self.a, self.b, self.c, self.d)]


@register("AG")
class construct_triangle(ConstructionRule):
    def __init__(self, a, b, c):
        self.a, self.b, self.c = a, b, c
        
    def constructed_points(self):
        return [self.a, self.b, self.c]


@register("AG")
class construct_triangle12(ConstructionRule):
    def __init__(self, a, b, c):
        self.a, self.b, self.c = a, b, c
        
    def constructed_points(self):
        return [self.a, self.b, self.c]
    
    def conclusions(self):
        return [Length(self.a, self.b) / Length(self.a, self.c) - 1 / 2]


@register("AG")
class construct_2l1c(ConstructionRule):
    def __init__(self, x, y, z, i, a, b, c, o):
        self.x, self.y, self.z, self.i, self.a, self.b, self.c, self.o = x, y, z, i, a, b, c, o
    
    def arguments(self):
        return [self.a, self.b, self.c, self.o]
        
    def constructed_points(self):
        return [self.x, self.y, self.z, self.i]
    
    def conditions(self):
        return [
            Length(self.o, self.a) - Length(self.o, self.b),
            NotCollinear(self.a, self.b, self.c),
        ]

    def conclusions(self):
        return [
            Collinear(self.x, self.a, self.c),
            Collinear(self.y, self.b, self.c),
            Length(self.o, self.a) - Length(self.o, self.z),
            Collinear(self.i, self.o, self.z),
            Length(self.i, self.x) - Length(self.i, self.y),
            Length(self.i, self.y) - Length(self.i, self.z),
            Perpendicular(self.i, self.x, self.a, self.c),
            Perpendicular(self.i, self.y, self.b, self.c),
        ]


@register("AG")
class construct_e5128(ConstructionRule):
    def __init__(self, x, y, a, b, c, d):
        self.x, self.y, self.a, self.b, self.c, self.d = x, y, a, b, c, d
    
    def arguments(self):
        return [self.a, self.b, self.c, self.d]
        
    def constructed_points(self):
        return [self.x, self.y]
    
    def conditions(self):
        return [
            Length(self.c, self.b) - Length(self.c, self.d),
            Perpendicular(self.b, self.c, self.b, self.a),
        ]

    def conclusions(self):
        return [
            Length(self.c, self.b) - Length(self.c, self.x),
            Collinear(self.y, self.a, self.b),
            Collinear(self.x, self.y, self.d),
            Angle(self.b, self.a, self.d) - Angle(self.a, self.x, self.y),
        ]


@register("AG")
class construct_3peq(ConstructionRule):
    def __init__(self, x, y, z, a, b, c):
        self.x, self.y, self.z, self.a, self.b, self.c = x, y, z, a, b, c
    
    def arguments(self):
        return [self.a, self.b, self.c]
        
    def constructed_points(self):
        return [self.x, self.y, self.z]
    
    def conditions(self):
        return [NotCollinear(self.a, self.b, self.c)]

    def conclusions(self):
        return [
            Collinear(self.z, self.b, self.c),
            Collinear(self.x, self.a, self.b),
            Collinear(self.y, self.a, self.c),
            Collinear(self.x, self.y, self.z),
            Length(self.z, self.x) - Length(self.z, self.y),
        ]


@register("AG")
class construct_trisect(ConstructionRule):
    def __init__(self, x, y, a, b, c):
        self.x, self.y, self.a, self.b, self.c = x, y, a, b, c
    
    def arguments(self):
        return [self.a, self.b, self.c]
        
    def constructed_points(self):
        return [self.x, self.y]
    
    def conditions(self):
        return [NotCollinear(self.a, self.b, self.c)]

    def conclusions(self):
        return [
            Angle(self.a, self.b, self.x) - Angle(self.x, self.b, self.y),
            Angle(self.x, self.b, self.y) - Angle(self.y, self.b, self.c),
            Collinear(self.x, self.a, self.c),
            Collinear(self.y, self.a, self.c),
        ]


@register("AG")
class construct_trisegment(ConstructionRule):
    def __init__(self, x, y, a, b):
        self.x, self.y, self.a, self.b = x, y, a, b
    
    def arguments(self):
        return [self.a, self.b]
        
    def constructed_points(self):
        return [self.x, self.y]
    
    def conditions(self):
        return [Different(self.a, self.b)]

    def conclusions(self):
        return [
            Length(self.a, self.x) - Length(self.x, self.y),
            Length(self.x, self.y) - Length(self.y, self.b),
            Collinear(self.x, self.a, self.b),
            Collinear(self.y, self.a, self.b),
        ]


@register("AG")
class construct_on_dia(ConstructionRule):
    def __init__(self, x, a, b):
        self.x, self.a, self.b = x, a, b
    
    def arguments(self):
        return [self.a, self.b]
        
    def constructed_points(self):
        return [self.x]
        
    def conditions(self):
        return [Different(self.a, self.b)]

    def conclusions(self):
        return [Perpendicular(self.x, self.a, self.x, self.b)]


@register("AG")
class construct_ieq_triangle(ConstructionRule):
    def __init__(self, a, b, c):
        self.a, self.b, self.c = a, b, c
        
    def constructed_points(self):
        return [self.a, self.b, self.c]
    
    def conclusions(self):
        return [
            Length(self.a, self.b) - Length(self.b, self.c),
            Length(self.b, self.c) - Length(self.c, self.a),
            Angle(self.b, self.a, self.c) - Angle(self.a, self.c, self.b),
            Angle(self.a, self.c, self.b) - Angle(self.c, self.b, self.a),
        ]


@register("AG")
class construct_on_opline(ConstructionRule):
    def __init__(self, x, a, b):
        self.x, self.a, self.b = x, a, b
    
    def arguments(self):
        return [self.a, self.b]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [Different(self.a, self.b)]

    def conclusions(self):
        return [Collinear(self.x, self.a, self.b)]


class construct_cc_tangent0(ConstructionRule):
    def __init__(self, x, y, o, a, w, b):
        self.x, self.y, self.o, self.a, self.w, self.b = x, y, o, a, w, b
    
    def arguments(self):
        return [self.o, self.a, self.w, self.b]
        
    def constructed_points(self):
        return [self.x, self.y]
    
    def conditions(self):
        return [
            Different(self.o, self.a),
            Different(self.w, self.b),
            Different(self.o, self.w),
        ]

    def conclusions(self):
        return [
            Length(self.o, self.x) - Length(self.o, self.a),
            Length(self.w, self.y) - Length(self.w, self.b),
            Perpendicular(self.x, self.o, self.x, self.y),
            Perpendicular(self.y, self.w, self.y, self.x)
        ]
        

class construct_cc_tangent(ConstructionRule):
    def __init__(self, x, y, z, i, o, a, w, b):
        self.x, self.y, self.z, self.i, self.o, self.a, self.w, self.b = x, y, z, i, o, a, w, b
    
    def arguments(self):
        return [self.o, self.a, self.w, self.b]
        
    def constructed_points(self):
        return [self.x, self.y, self.z, self.i]
    
    def conditions(self):
        return [
            Different(self.o, self.a),
            Different(self.w, self.b),
            Different(self.o, self.w),
        ]

    def conclusions(self):
        return [
            Length(self.o, self.x) - Length(self.o, self.a),
            Length(self.w, self.y) - Length(self.w, self.b),
            Perpendicular(self.x, self.o, self.x, self.y),
            Perpendicular(self.y, self.w, self.y, self.x),
            Length(self.o, self.z) - Length(self.o, self.a),
            Length(self.w, self.i) - Length(self.w, self.b),
            Perpendicular(self.z, self.o, self.z, self.i),
            Perpendicular(self.i, self.w, self.i, self.z),
        ]


@register("AG")
class construct_eqangle3(ConstructionRule):
    def __init__(self, x, a, b, d, e, f):
        self.x, self.a, self.b, self.d, self.e, self.f = x, a, b, d, e, f
    
    def arguments(self):
        return [self.a, self.b, self.d, self.e, self.f]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [
            NotCollinear(self.d, self.e, self.f),
            Different(self.a, self.b),
            Different(self.d, self.e),
            Different(self.e, self.f),
        ]

    def conclusions(self):
        return [Angle(self.a, self.x, self.b) - Angle(self.e, self.d, self.f)]


@register("AG")
class construct_tangent(ConstructionRule):
    def __init__(self, x, y, a, o, b):
        self.x, self.y, self.a, self.o, self.b = x, y, a, o, b
    
    def arguments(self):
        return [self.a, self.o, self.b]
        
    def constructed_points(self):
        return [self.x, self.y]
    
    def conditions(self):
        return [
            Different(self.o, self.a),
            Different(self.o, self.b),
            Different(self.a, self.b),
        ]

    def conclusions(self):
        return [
            Length(self.o, self.x) - Length(self.o, self.b),
            Perpendicular(self.a, self.x, self.o, self.x),
            Length(self.o, self.y) - Length(self.o, self.b),
            Perpendicular(self.a, self.y, self.o, self.y),
        ]


@register("AG")
class construct_on_circum(ConstructionRule):
    def __init__(self, x, a, b, c):
        self.x, self.a, self.b, self.c = x, a, b, c
    
    def arguments(self):
        return [self.a, self.b, self.c]
        
    def constructed_points(self):
        return [self.x]
        
    def conditions(self):
        return [NotCollinear(self.a, self.b, self.c)]
    
    def conclusions(self):
        return [Concyclic(self.a, self.b, self.c, self.x)]


class construct_sameside(ConstructionRule):
    def __init__(self, x, a, b, c):
        self.x, self.a, self.b, self.c = x, a, b, c
    
    def arguments(self):
        return [self.a, self.b, self.c]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [NotCollinear(self.a, self.b, self.c)]

    def conclusions(self):
        return [SameSide(self.x, self.a, self.b, self.c)]


class construct_opposingsides(ConstructionRule):
    def __init__(self, x, a, b, c):
        self.x, self.a, self.b, self.c = x, a, b, c
    
    def arguments(self):
        return [self.a, self.b, self.c]
        
    def constructed_points(self):
        return [self.x]

    def conditions(self):
        return [NotCollinear(self.a, self.b, self.c)]

    def conclusions(self):
        return [OppositeSide(self.x, self.a, self.b, self.c)]


"""Parsing utilities for benchmark text descriptions into constructions/goals."""

from collections import Counter

from pyeuclid.formalization.relation import *
from pyeuclid.formalization.construction_rule import *
from pyeuclid.formalization.utils import is_float


def get_constructions_list_from_text(text):
    """Parse the constructions section of a text instance into rule objects.

    Args:
        text (str): Full benchmark line containing constructions and goal separated by ' ? '.

    Returns:
        list[list[ConstructionRule]]: Nested list of construction batches.
    """
    parts = text.split(' ? ')
    constructions_text_list = parts[0].split('; ')
    constructions_list = []
    
    for constructions_text in constructions_text_list:
        constructions_text = constructions_text.split(' = ')[1]
        construction_text_list = constructions_text.split(', ')
        constructions = []
        for construction_text in construction_text_list:
            construction_text = construction_text.split(' ')
            rule_name = construction_text[0]
            arg_names = [name.replace('_', '') for name in construction_text[1:]]
            rule = globals()['construct_'+rule_name]
            args = [float(arg_name) if is_float(arg_name) else Point(arg_name) for arg_name in arg_names]
            construction = rule(*args)
            constructions.append(construction)
        constructions_list.append(constructions)
    
    return constructions_list

def get_goal_from_text(text):
    """Parse the goal portion of a text instance into a Relation or expression.

    Args:
        text (str): Full benchmark line containing constructions and goal.

    Returns:
        Relation | sympy.Expr | tuple[Relation, Relation] | None: Parsed goal or None.
    """
    parts = text.split(' ? ')
    goal_text = parts[1] if len(parts) > 1 else None
    goal = None
    if goal_text:
        goal_text = goal_text.split(' ')
        goal_name = goal_text[0]
        arg_names = [name.replace('_', '') for name in goal_text[1:]]
        args = [Point(arg_name) for arg_name in arg_names]
        if goal_name == 'cong':
            goal = Length(*args[:2]) - Length(*args[2:])
        elif goal_name == 'cyclic':
            goal = Concyclic(*args)
        elif goal_name == 'coll':
            goal = Collinear(*args)
        elif goal_name == 'perp':
            goal = Perpendicular(*args)
        elif goal_name == 'para':
            goal = Parallel(*args)
        elif goal_name == 'eqratio':
            goal = Length(*args[:2])/Length(*args[2:4]) - Length(*args[4:6])/Length(*args[6:8])
        elif goal_name == 'eqangle':
            def extract_angle(points):
                count = Counter(points)
                repeating = next(p for p, c in count.items() if c == 2)
                singles = [p for p, c in count.items() if c == 1]
                return singles[0], repeating, singles[1]
            angle1 = Angle(*extract_angle(args[:4]))
            angle2 = Angle(*extract_angle(args[4:]))
            # The goal may involve either equal angles or supplementary angles
            goal = (angle1 - angle2, angle1 + angle2 - pi)
        elif goal_name == 'midp':
            goal = Midpoint(*args)
        elif goal_name == 'simtri':
            goal = Similar(*args)
        elif goal_name == 'contri':
            goal = Congruent(*args)
    
    return goal


def parse_texts_from_file(file_name):
    """Load every other line from a benchmark file as a problem description.

    Args:
        file_name (str): Path to benchmark text file.

    Returns:
        list[str]: Problem description strings.
    """
    with open(file_name, "r") as f:
        lines = f.readlines()
    
    texts = [lines[i].strip() for i in range(1, len(lines), 2)]
    return texts
            


import re
import sympy
from typing import List
from pathlib import Path

ROOT_DIR = Path(__file__).parents[2]
MAX_DIAGRAM_ATTEMPTS = 100


def sort_points(*points):
    return sorted(points, key=lambda i: i.name)

def sort_cyclic_points(*points):
    min_index = min(range(len(points)), key=lambda i: points[i].name)
    if str(points[(min_index+1) % len(points)]) > str(points[(min_index-1) % len(points)]):
        remaining_list = list(points[min_index:] + points[:min_index])[1:]
        return [points[min_index]] + remaining_list[::-1]
    else:
        return points[min_index:] + points[:min_index]


def compare_names(g1, g2):
    assert len(g1) == len(g2)
    for a, b in zip(g1, g2):
        if a.name != b.name:
            return a.name < b.name
    return True


def get_point_mapping(g1, g2):
    mapping = {}
    for p1, p2 in zip(g1, g2):
        mapping[p1] = p2
        mapping[p2] = p1
    return mapping


def sort_point_groups(g1, g2, mapping=None):
    if not compare_names(g1, g2):
        g1, g2 = g2, g1
    
    if mapping:
        g2 = [mapping[p] for p in g1]
    
    return g1 + g2


def expand_definition(relation):
    if not type(relation) in (tuple, list):
        relation = relation,
    lst = []
    for prop in relation:
        if hasattr(prop, "definition"):
            for item in prop.definition():
                lst += expand_definition(item)
        else:
            lst += prop,
    return lst


class UnionFind:
    def __init__(self):
        self.parent = {}
        self.rank = {}

    def find(self, item):
        neg = getattr(item, "neg", False)
        if neg:
            item = -item
        if not item in self.parent:
            self.add(item)
            return -item if neg else item
        if self.parent[item] != item:
            self.parent[item] = self.find(self.parent[item])
        return -self.parent[item] if neg else self.parent[item]

    def union(self, item1, item2):
        if (type(item1).__name__ == 'Length' and type(item1).__name__ == 'Angle') or (type(item1).__name__ == 'Angle' and type(item1).__name__ == 'Length'):
            breakpoint()
            assert False
        root1 = self.find(item1)
        root2 = self.find(item2)
        neg = getattr(root1, "neg", False) ^ getattr(root2, "neg", False)
        if getattr(root1, "neg", False):
            root1 = - root1
        if getattr(root2, "neg", False):
            root2 = - root2
        if root1 != root2:
            if self.rank[root1] > self.rank[root2]:
                self.parent[root2] = -root1 if neg else root1
            elif self.rank[root1] < self.rank[root2]:
                self.parent[root1] = -root2 if neg else root2
            else:
                self.parent[root2] = -root1 if neg else root1
                self.rank[root1] += 1

    def add(self, item):
        if getattr(item, "neg", False):
            item = -item
        if item not in self.parent:
            self.parent[item] = item
            self.rank[item] = 0

    def equivalence_classes(self):
        dic = {}
        for item in self.parent:
            root = self.find(item)
            if root in dic:
                dic[root].append(item)
            else:
                dic[root] = [item]
        return dic

    def merge_eq(uf, v1, v2):
        uf.add(v1)
        uf.add(v2)
        uf.union(v1, v2)
        for v in [v1, v2]:
            if uf.parent[v] != v:
                v.rep_by = uf.parent[v]
                setattr(v.rep_by, "rep_by", None)


class Traced():
    def __init__(self, expr, depth=0, sources=[]):
        if isinstance(expr, Traced):
            sources = expr.sources
            depth = expr.depth
            expr = expr.expr
        self.expr = expr
        self.symbol = None
        self.redundant = False
        self.sources = sources
        self.depth = max([depth] + [getattr(item, "depth", 0) for item in self.sources])
        for key in ("free_symbols", "args"):
            setattr(self, key, getattr(self.expr, key))
    
    def subs(self, key, value):
        assert isinstance(value, Traced)
        if len(self.sources)>0 and isinstance(self.sources[0], Traced):
            sources = [item for item in self.sources] + [value]
        else:
            sources = [self, value]
        value.symbol = key
        expr = self.expr.subs(key, value.expr)
        other = Traced(expr, sources=sources)
        other.symbol = self.symbol
        return other
    
    def __str__(self):
        if not self.symbol is None:
            return str(self.symbol - self.expr)
        return str(self.expr)
    def __repr__(self):
        return str(self)
    def __eq__(self, other):
        return hash(self) == hash(other)
    def __hash__(self):
        rep = f"{self}@{self.depth}"
        rep += " ".join([str(hash(item)) for item in self.sources])
        return hash(rep)

        
def infer_eq_types(eq, var_types):
    eq_types = set()
    for symbol in eq.free_symbols:
        if "Length" in str(symbol):
            eq_types.add("Length")
        elif "Angle" in str(symbol):
            eq_types.add("Angle")
        elif symbol in var_types and not var_types[symbol] is None:
            eq_types.add(var_types[symbol])
    return eq_types
    
    
def classify_equations(equations: List[Traced], var_types):
    angle_linear, length_linear, length_ratio, others = [], [], [], []
    cnst = r"(\d+|\d+\.\d*|pi)"
    cnst = f"{cnst}(\\*{cnst})*(/{cnst})*"
    length = r"(length\w+\d*|variable\w+\d*)"
    angle = r"(angle\w+\d*|variable\w+\d*)"
    length_mono = f"({cnst}\\*)?{length}(/{cnst})?"
    angle_mono = f"({cnst}\\*)?{angle}(/{cnst})?"
    length_mono = f"({length_mono}|{cnst})"
    angle_mono = f"({angle_mono}|{cnst})"
    length_ratio_pattern = re.compile(
        f"^-?{length_mono}([\\*/]{length_mono})* [+-] {length_mono}([\\*/]{length_mono})*$")
    length_linear_pattern = re.compile(
        f"^-?{length_mono}( [+-] {length_mono})+$")
    angle_linear_pattern = re.compile(
        f"^-?{angle_mono}( [+-] {angle_mono})+$")
    for i, eq in enumerate(equations):
        tmp = str(eq.expr.expand()).lower()
        eq_types = infer_eq_types(eq, var_types)
        if length_ratio_pattern.match(tmp):
            # length ratio has higher priority than length linear
            # eqratio, length eq const, eqlength, linear equations involving length and variables
            if "Length" in eq_types:
                length_ratio.append(eq)
            elif "Angle" in eq_types: # such as Variable_angle_2 - piVariable_x/90
                angle_linear.append(eq)
            else:
                others.append(eq)
        elif angle_linear_pattern.match(tmp) or length_linear_pattern.match(tmp):
            # eqangle, angle eq const, angle sum
            if len(eq_types) > 1:
                breakpoint()
                assert False
            if len(eq_types) == 0:
                others.append(eq)
            elif eq_types.pop() == "Angle":
                angle_linear.append(eq)
            else:
                length_linear.append(eq)
        else:
            others.append(eq)
    return angle_linear, length_linear, length_ratio, others

def is_float(s):
    try:
        float(s)
        return True
    except ValueError:
        return False


def parse_expression(expr):
    symbols = {'Angle': [], 'Length': []}
    symbol_names = {'Angle': [], 'Length': []}
    
    for arg in expr.free_symbols:
        if arg.is_Symbol:
            match1 = re.match(r'Angle_(\w+)_(\w+)_(\w+)', arg.name)
            match2 = re.match(r'Length_(\w+)_(\w+)', arg.name)
        if match1:
            symbols['Angle'].append(arg)
            symbol_names['Angle'].append(list(match1.groups()))
        if match2:
            symbols['Length'].append(arg)
            symbol_names['Length'].append(list(match2.groups()))            
            
    return symbols, symbol_names

eps = 1e-3
def is_small(x):
    if len(x.free_symbols) > 0:
        return False
    if hasattr(x, "evalf"):
        x = x.evalf()
    try:
        return abs(x) < eps
    except:
        assert False

def check_equalities(equalities):
    if not type(equalities) in (tuple, list):
        equalities = [equalities]
    for cond in equalities:
        if not (isinstance(cond, sympy.logic.boolalg.BooleanTrue) or isinstance(cond, sympy.core.numbers.Zero) or is_small(cond)):
            return False
    return True


import math
import gurobipy as gp
import numpy as np
import sympy
from gurobipy import GRB
from pyeuclid.formalization.relation import *
from pyeuclid.formalization.utils import *
from pyeuclid.engine.inference_rule import *


class ProofGenerator:
    def __init__(self, state):
        self.state = state
        self.proof = ""
    
    def show_proof(self):
        print(self.proof)
    
    def traceback(self, augmented_A, e) -> list[str]:
        m, n = augmented_A.shape
        e = e[0]
        n = n-1

        model = gp.Model()
        model.setParam('OutputFlag', 0)

        x = model.addVars(m, lb=-GRB.INFINITY, ub=GRB.INFINITY, vtype=GRB.CONTINUOUS, name="x")
        z = model.addVars(m, vtype=GRB.BINARY, name="z")

        model.setObjective(gp.quicksum(z[i] for i in range(m)), GRB.MINIMIZE)

        for i in range(n + 1):
            model.addConstr(gp.quicksum(augmented_A[j, i] * x[j] for j in range(m)) == e[i])

        M = 1e6
        for i in range(m):
            model.addConstr(x[i] <= M * z[i])
            model.addConstr(x[i] >= -M * z[i])

        model.optimize()

        assert model.status == GRB.OPTIMAL        
        return [i for i in range(m) if z[i].x > 0]
    
    def vectorize(self, equations, variables, source):
        A = np.zeros(shape=(len(equations), len(variables)), dtype=np.float64)
        b = np.zeros(shape=(len(equations), 1), dtype=np.float64)
        if source in ("angle_linear", "length_linear"):
            for i, eqn in enumerate(equations):
                eqn = sympy.expand(eqn)
                assert isinstance(eqn, sympy.core.add.Add) or isinstance(eqn, sympy.core.symbol.Symbol)

                for add_arg in eqn.args:
                    if len(add_arg.args) == 0:
                        mul_args = [add_arg]
                    else:
                        assert isinstance(add_arg, sympy.core.mul.Mul)
                        mul_args = add_arg.args
                    factors = [item for item in mul_args if len(
                        item.free_symbols) == 0]
                    factor = sympy.core.mul.Mul(*factors)
                    symbols = [item for item in mul_args if len(
                        item.free_symbols) > 0]
                    if len(symbols) == 0:
                        b[i, 0] = factor.evalf()
                    else:
                        A[i, variables[symbols[0]]] = factor.evalf()
        else:
            assert source == "length_ratio"  # length=const or eqlength or eqlength ratio or lengthratio=const
            for i, eqn in enumerate(equations):
                if isinstance(eqn, sympy.core.add.Add):
                    if len(eqn.args) > 2:
                        breakpoint()
                        assert False
                    add_args = eqn.args
                else:
                    add_args = [eqn]
                for j, add_arg in enumerate(add_args):
                    if len(add_arg.args) > 0:
                        mul_args = add_arg.args
                    else:
                        mul_args = [add_arg]
                    for mul_arg in mul_args:
                        factor = (-1)**(j)
                        if isinstance(mul_arg, sympy.core.power.Pow):
                            factor *= mul_arg.args[1]
                            mul_arg = mul_arg.args[0]
                        if len(mul_arg.free_symbols) == 0:
                            b[i, 0] += factor * math.log(abs(mul_arg))
                        else:
                            symbol = list(mul_arg.free_symbols)[0]
                            A[i, variables[symbol]] += factor
        return np.concat([A, b], axis=1)


    def find_conditions(self, equations: list[Traced], conclusion, source):
        angle_linear, length_linear, length_ratio, others = classify_equations(equations, self.state.var_types)
        """Given sympified equations and conclusions, return a list of necessary conditions"""
        def try_find(equations, conclusion):
            variables = set()
            for eqn in equations:
                variables = variables.union(eqn.free_symbols)
            variables = {item: i for i, item in enumerate(list(variables))}
            mat = self.vectorize([item.expr for item in equations], variables, source)
            eq = self.vectorize([conclusion], variables, source)
            deps = self.traceback(mat, eq)
            return [equations[i] for i in deps]
        if source == "angle_linear":
            equations = angle_linear
        elif source == "length_linear":
            equations = length_linear
        else:
            assert source == "length_ratio"
            equations = length_ratio
        return try_find(equations, conclusion)
    
    def run(self, node, visited=set([]), depth=None, verbose=False, root=True):
        if root:
            depth = self.state.current_depth
            visited = set([])
        elif depth is None:
            depth = node.depth
        # self.logger.debug(f"{node}@{depth}: {getattr(node, 'sources', None)}")
        def format_proof(proof_dict, conclusion):
            proof_steps = {}
            visited = set()
            step_counter = 1

            def format_conditions(condition, proof_steps, theorem):
                s = []
                for condition in conditions:
                    if condition in proof_steps:
                        s.append(f"{condition}({proof_steps[condition][0]})")
                    else:
                        s.append(f"{condition}")
                if theorem is None:
                    return " &\n".join(s)
                return " &\n".join(s) + f"({theorem})"
            def search(node):  # root-last traversal
                nonlocal step_counter
                if node in visited or node not in proof_dict or proof_dict[node] is None:
                    return
                visited.add(node)
                conditions = proof_dict[node]
                theorem = None
                while len(conditions) == 1 and conditions[0] in proof_dict: # collapse single-condition inferences
                    if isinstance(conditions[0], InferenceRule) and not type(conditions[0]) in inference_rule_sets["ex"]:
                        theorem = conditions[0]
                    conditions = proof_dict[conditions[0]]
                for condition in conditions:
                    if condition is not None:
                        search(condition)
                if all([type(item) in (Collinear, Between, SameSide)for item in conditions]):
                    return
                if type(node) in (Traced, sympy.core.add.Add):
                    for item in visited:
                        if not item is node and type(item) in (Traced, sympy.core.add.Add):
                            if getattr(node, "expr", node) - getattr(item, "expr", item) == 0:
                                return
                            if getattr(node, "expr", node) + getattr(item, "expr", item) == 0:
                                return
                proof_steps[node] = (step_counter, conditions, theorem)
                step_counter += 1
            search(conclusion)
            self.proof += "* Proof steps:\n"
            lst = [(key, value) for key, value in proof_steps.items()]
            lst.sort(key=lambda x: x[1][0])
            last = -1
            for node, (step_number, conditions, theorem) in lst:
                if step_number == last:
                    continue
                last = step_number
                self.proof += f"{step_number:03}. {format_conditions(node, proof_steps, theorem)} ⇒ {node}\n"
        if node in visited:
            return {}
        if isinstance(node, InferenceRule):
            visited.add(node)
            conds = [item for item in node.condition() if type(item)
                     not in (Equal, Lt) and not item == 0]
            result = {}
            result[node] = conds
            for cond in conds:
                result.update(self.run(cond, visited, depth=depth-1, root=False))
        elif isinstance(node, Relation):
            visited.add(node)
            if type(node) in (Between, SameSide, Lt, Equal) or type(node) == Collinear and (node.p1 == node.p2 or node.p2 == node.p3 or node.p3 == node.p1):
                return {}
            if hasattr(node, "definition"):
                defs = node.definition()
                result = {node: defs}
                for cond in defs:
                    result.update(self.run(cond, visited, depth=depth, root=False))
                return result
            result = {}
            for tmp in self.state.relations:
                if tmp == node:
                    if hasattr(tmp, "source"):
                        source = tmp.source
                        result = {node: [source]}
                        result.update(self.run(source, visited, root=False))
                    else:
                        result = {node: []}
                    break
            assert len(result) > 0, f"{node} is not proved"
        else:
            if isinstance(node, Traced):
                sources = node.sources
                if len(sources) == 0: # initial conditions
                    visited.add(node)
                elif isinstance(sources[0], str):
                    # backtrace linear systems
                    equations = [item for item in self.state.equations if item.depth <= node.depth]
                    if not node.symbol is None:
                        expr = node.symbol - node.expr
                    else:
                        expr = node.expr
                    conditions = self.find_conditions(equations, expr, sources[0])
                    if not conditions:
                        breakpoint()
                        assert False
                    sources = conditions
                else:
                    visited.add(node)
            else:
                assert isinstance(node, sympy.core.expr.Expr)
                sources = []
                solved_vars = self.state.solutions[min(int(depth), len(self.state.solutions)-1)]
                for symbol in node.free_symbols:
                    if not symbol in solved_vars:
                        continue  # free vars
                    expr = symbol-solved_vars[symbol].expr
                    expr = Traced(expr, sources=solved_vars[symbol].sources, depth=int(depth))
                    sources.append(expr)
                if isinstance(node, sympy.core.symbol.Symbol):
                    node = node - solved_vars[node].expr
            result = {node: sources}
            for item in sources:
                result.update(self.run(item, visited, root=False))
        if root:
            format_proof(result, node)
        return result
    
    def generate_proof(self):
        self.run(self.state.goal)


"""Numerical diagram construction and validation for geometric problems."""
# mypy: ignore-errors

from __future__ import annotations

import os
import pickle
import hashlib

from matplotlib import pyplot as plt  # type: ignore[import]

from pyeuclid.formalization.construction_rule import *
from pyeuclid.formalization.numericals import *
from pyeuclid.formalization.utils import *


def hash_constructions_list(constructions_list):
    s = ",".join(str(c) for constructions in constructions_list for c in constructions)
    return hashlib.md5(s.encode('utf-8')).hexdigest()


class Diagram:    
    def __new__(cls, constructions_list:list[list[ConstructionRule]]=None, save_path=None, cache_folder=os.path.join(ROOT_DIR, 'cache'), resample=False):
        """Load from cache if available, otherwise construct a new diagram instance."""
        if not resample and cache_folder is not None:
            if not os.path.exists(cache_folder):
                os.makedirs(cache_folder)
            
            if constructions_list is not None:
                file_name = f"{hash_constructions_list(constructions_list)}.pkl"
                file_path = os.path.join(cache_folder, file_name)
                try:
                    if os.path.exists(file_path):
                        with open(file_path, 'rb') as f:
                            instance = pickle.load(f)
                            instance.save_path = save_path
                            instance.save_diagram()
                            return instance
                except:
                    pass
        
        instance = super().__new__(cls)
        return instance
    
    def __init__(self, constructions_list:list[list[ConstructionRule]]=None, save_path=None, cache_folder=os.path.join(ROOT_DIR, 'cache'), resample=False):
        if hasattr(self, 'cache_folder'):
            return
        
        self.points = []
        self.segments = []
        self.circles = []
        
        self.name2point = {}
        self.point2name = {}
        
        self.fig, self.ax = None, None
        
        self.constructions_list = constructions_list
        self.save_path = save_path
        self.cache_folder = cache_folder
        
        if constructions_list is not None:                
            self.construct_diagram()
            
    def clear(self):
        """Reset all stored points, segments, circles, and name mappings."""
        self.points.clear()
        self.segments.clear()
        self.circles.clear()
        
        self.name2point.clear()
        self.point2name.clear()
        
    def show(self):
        """Render the diagram with matplotlib."""
        self.draw_diagram(show=True)
        
    def save_to_cache(self):
        """Persist the diagram to cache if caching is enabled."""
        if self.cache_folder is not None:
            file_name = f"{hash_constructions_list(self.constructions_list)}.pkl"
            file_path = os.path.join(self.cache_folder, file_name)
            with open(file_path, 'wb') as f:
                pickle.dump(self, f)
    
    def add_constructions(self, constructions):
        """Add a new batch of constructions, retrying if degeneracy occurs."""
        for _ in range(MAX_DIAGRAM_ATTEMPTS):
            try:
                self.construct(constructions)
                self.constructions_list.append(constructions)
                return
            except:
                continue
        
        print(f"Failed to add the constructions after {MAX_DIAGRAM_ATTEMPTS} attempts.")
        raise Exception()
            
    def construct_diagram(self):
        """Construct the full diagram from all construction batches, with retries."""
        for _ in range(MAX_DIAGRAM_ATTEMPTS):
            try:
                self.clear()
                for constructions in self.constructions_list:
                    self.construct(constructions)
                self.draw_diagram()
                self.save_to_cache()
                return
            except:
                continue
        
        print(f"Failed to construct a diagram after {MAX_DIAGRAM_ATTEMPTS} attempts.")
        raise Exception()
            
    def construct(self, constructions: list[ConstructionRule]):
        """Apply a single batch of construction rules to extend the diagram."""
        constructed_points = constructions[0].constructed_points()
        if any(construction.constructed_points() != constructed_points for construction in constructions[1:]):
            raise Exception()

        to_be_intersected = []
        for construction in constructions:
            # print(construction.__class__.__name__ + '('+','.join([str(name) for name in construction.arguments()])+')')
            # for c in construction.conditions:
            #     if not self.numerical_check(c):
            #         raise Exception()
            
            to_be_intersected += self.sketch(construction)
                
        new_points = self.reduce(to_be_intersected, self.points)
        
        if check_too_close(new_points, self.points):
            raise Exception()
        
        if check_too_far(new_points, self.points):
            raise Exception()
        
        self.points += new_points
        
        for p, np in zip(constructed_points, new_points):
            self.name2point[p.name] = np
            self.point2name[np] = p.name
        
        for construction in constructions:
            self.draw(new_points, construction)
            
    def numerical_check_goal(self, goal):
        """Check if the current diagram satisfies a goal relation/expression."""
        if isinstance(goal, tuple):
            for g in goal:
                if self.numerical_check(g):
                    return True, g
        else:
            if self.numerical_check(goal):
                return True, goal
        return False, goal
            
    def numerical_check(self, relation):
        """Numerically evaluate whether a relation/expression holds in the diagram."""
        if isinstance(relation, Relation):
            func = globals()['check_' + relation.__class__.__name__.lower()]
            args = [self.name2point[p.name] for p in relation.get_points()]
            return func(args)
        else:
            symbol_to_value = {}
            symbols, symbol_names = parse_expression(relation)
            
            for angle_symbol, angle_name in zip(symbols['Angle'], symbol_names['Angle']):
                angle_value = calculate_angle(*[self.name2point[n] for n in angle_name])
                symbol_to_value[angle_symbol] = angle_value
            
            for length_symbol, length_name in zip(symbols['Length'], symbol_names['Length']):
                length_value = calculate_length(*[self.name2point[n] for n in length_name])
                symbol_to_value[length_symbol] = length_value
            
            evaluated_expr = relation.subs(symbol_to_value)
            if close_enough(float(evaluated_expr.evalf()), 0):
                return True
            else:
                return False

    def sketch(self, construction):
        func = getattr(self, 'sketch_' + construction.__class__.__name__[10:])
        args = [arg if isinstance(arg, float) else self.name2point[arg.name] for arg in construction.arguments()]
        result = func(*args)
        if isinstance(result, list):
            return result
        else:
            return [result]
        
    def sketch_angle_bisector(self, *args: list[Point]) -> Ray:
        """Ray that bisects angle ABC."""
        a, b, c = args
        dist_ab = a.distance(b)
        dist_bc = b.distance(c)
        x = b + (c - b) * (dist_ab / dist_bc)
        m = (a + x) * 0.5
        return Ray(b, m)
    
    def sketch_angle_mirror(self, *args: list[Point]) -> Ray:
        """Mirror of ray BA across BC."""
        a, b, c = args
        ab = a - b
        cb = c - b

        dist_ab = a.distance(b)
        ang_ab = np.arctan2(ab.y / dist_ab, ab.x / dist_ab)
        dist_cb = c.distance(b)
        ang_bc = np.arctan2(cb.y / dist_cb, cb.x / dist_cb)

        ang_bx = 2 * ang_bc - ang_ab
        x = b + Point(np.cos(ang_bx), np.sin(ang_bx))
        return Ray(b, x)
    
    def sketch_circle(self, *args: list[Point]) -> Point:
        """Center of circle through three points."""
        a, b, c = args
        l1 = perpendicular_bisector(a, b)
        l2 = perpendicular_bisector(b, c)
        x = line_line_intersection(l1, l2)
        return x
    
    def sketch_circumcenter(self, *args: list[Point]) -> Point:
        """Circumcenter of triangle ABC."""
        a, b, c = args
        l1 = perpendicular_bisector(a, b)
        l2 = perpendicular_bisector(b, c)
        x = line_line_intersection(l1, l2)
        return x
    
    def sketch_eq_quadrangle(self, *args: list[Point]) -> list[Point]:
        """Randomly sample a quadrilateral with opposite sides equal."""
        a = Point(0.0, 0.0)
        b = Point(1.0, 0.0)

        length = np.random.uniform(0.5, 2.0)
        ang = np.random.uniform(np.pi / 3, np.pi * 2 / 3)
        d = head_from(a, ang, length)

        ang = ang_of(b, d)
        ang = np.random.uniform(ang / 10, ang / 9)
        c = head_from(b, ang, length)
        a, b, c, d = random_rfss(a, b, c, d)
        return [a, b, c, d]
        
    def sketch_eq_trapezoid(self, *args: list[Point]) -> list[Point]:
        """Randomly sample an isosceles trapezoid."""
        a = Point(0.0, 0.0)
        b = Point(1.0, 0.0)
        l = unif(0.5, 2.0)

        height = unif(0.5, 2.0)
        c = Point(0.5 + l / 2.0, height)
        d = Point(0.5 - l / 2.0, height)

        a, b, c, d = random_rfss(a, b, c, d)
        return [a, b, c, d]
    
    def sketch_eq_triangle(self, *args: list[Point]) -> list[Circle]:
        """Circles defining an equilateral triangle on BC."""
        b, c = args
        return [Circle(center=b, radius=b.distance(c)), Circle(center=c, radius=b.distance(c))]
    
    def sketch_eqangle2(self, *args: list[Point]) -> Point:
        """Point X such that angle ABX equals angle XCB."""
        a, b, c = args
        
        ba = b.distance(a)
        bc = b.distance(c)
        l = ba * ba / bc

        if unif(0.0, 1.0) < 0.5:
            be = min(l, bc)
            be = unif(be * 0.1, be * 0.9)
        else:
            be = max(l, bc)
            be = unif(be * 1.1, be * 1.5)

        e = b + (c - b) * (be / bc)
        y = b + (a - b) * (be / l)
        return line_line_intersection(Line(c, y), Line(a, e))
    
    def sketch_eqdia_quadrangle(self, *args) -> list[Point]:
        """Quadrilateral with equal diagonals."""
        m = unif(0.3, 0.7)
        n = unif(0.3, 0.7)
        a = Point(-m, 0.0)
        c = Point(1 - m, 0.0)
        b = Point(0.0, -n)
        d = Point(0.0, 1 - n)

        ang = unif(-0.25 * np.pi, 0.25 * np.pi)
        sin, cos = np.sin(ang), np.cos(ang)
        b = b.rotate(sin, cos)
        d = d.rotate(sin, cos)
        a, b, c, d = random_rfss(a, b, c, d)
        return [a, b, c, d]
        
    def sketch_eqdistance(self, *args) -> Circle:
        """Circle centered at A with radius BC."""
        a, b, c = args
        return Circle(center=a, radius=b.distance(c))
    
    def sketch_eqdistance2(self, *args) -> Circle:
        """Circle centered at A with radius alpha*BC."""
        a, b, c, alpha = args
        return Circle(center=a, radius=alpha*b.distance(c))
    
    def sketch_eqdistance3(self, *args) -> Circle:
        """Circle centered at A with fixed radius alpha."""
        a, alpha = args
        return Circle(center=a, radius=alpha)
    
    def sketch_foot(self, *args) -> Point:
        """Foot of perpendicular from A to line BC."""
        a, b, c = args
        line_bc = Line(b, c)
        tline = a.perpendicular_line(line_bc)
        return line_line_intersection(tline, line_bc)
    
    def sketch_free(self, *args) -> Point:
        """Free point uniformly sampled in a box."""
        return Point(unif(-1, 1), unif(-1, 1))
    
    def sketch_incenter(self, *args) -> Point:
        """Incenter of triangle ABC."""
        a, b, c = args
        l1 = self.sketch_angle_bisector(a, b, c)
        l2 = self.sketch_angle_bisector(b, c, a)
        return line_line_intersection(l1, l2)
    
    def sketch_incenter2(self, *args) -> list[Point]:
        """Incenter plus touch points on each side."""
        a, b, c = args
        i = self.sketch_incenter(a, b, c)
        x = i.foot(Line(b, c))
        y = i.foot(Line(c, a))
        z = i.foot(Line(a, b))
        return [x, y, z, i]
    
    def sketch_excenter(self, *args) -> Point:
        """Excenter opposite B in triangle ABC."""
        a, b, c = args
        l1 = self.sketch_angle_bisector(b, a, c)
        l2 = self.sketch_angle_bisector(a, b, c).perpendicular_line(b)
        return line_line_intersection(l1, l2)
    
    def sketch_excenter2(self, *args) -> list[Point]:
        """Excenter plus touch points on extended sides."""
        a, b, c = args
        i = self.sketch_excenter(a, b, c)
        x = i.foot(Line(b, c))
        y = i.foot(Line(c, a))
        z = i.foot(Line(a, b))
        return [x, y, z, i]
    
    def sketch_centroid(self, *args) -> list[Point]:
        """Mid-segment points and centroid of triangle ABC."""
        a, b, c = args
        x = (b + c) * 0.5
        y = (c + a) * 0.5
        z = (a + b) * 0.5
        i = line_line_intersection(Line(a, x), Line(b, y))
        return [x, y, z, i]
    
    def sketch_intersection_cc(self, *args) -> list[Circle]:
        """Two circles centered at O and W through A."""
        o, w, a = args
        return [Circle(center=o, radius=o.distance(a)), Circle(center=w, radius=w.distance(a))]
    
    def sketch_intersection_lc(self, *args) -> list:
        """Line and circle defined by A,O,B for intersection."""
        a, o, b = args
        return [Line(b, a), Circle(center=o, radius=o.distance(b))]
    
    def sketch_intersection_ll(self, *args) -> Point:
        """Intersection of lines AB and CD."""
        a, b, c, d = args
        l1 = Line(a, b)
        l2 = Line(c, d)
        return line_line_intersection(l1, l2)
    
    def sketch_intersection_lp(self, *args) -> Point:
        a, b, c, m, n = args
        l1 = Line(a,b)
        l2 = self.sketch_on_pline(c, m, n)
        return line_line_intersection(l1, l2)
    
    def sketch_intersection_lt(self, *args) -> Point:
        a, b, c, d, e = args
        l1 = Line(a, b)
        l2 = self.sketch_on_tline(c, d, e)
        return line_line_intersection(l1, l2)
    
    def sketch_intersection_pp(self, *args) -> Point:
        a, b, c, d, e, f = args
        l1 = self.sketch_on_pline(a, b, c)
        l2 = self.sketch_on_pline(d, e, f)
        return line_line_intersection(l1, l2)
    
    def sketch_intersection_tt(self, *args) -> Point:
        a, b, c, d, e, f = args
        l1 = self.sketch_on_tline(a, b, c)
        l2 = self.sketch_on_tline(d, e, f)
        return line_line_intersection(l1, l2)
    
    def sketch_iso_triangle(self, *args) -> list[Point]:
        base = unif(0.5, 1.5)
        height = unif(0.5, 1.5)

        b = Point(-base / 2, 0.0)
        c = Point(base / 2, 0.0)
        a = Point(0.0, height)
        a, b, c = random_rfss(a, b, c)
        return [a, b, c]
    
    def sketch_lc_tangent(self, *args) -> Line:
        a, o = args
        return self.sketch_on_tline(a, a, o)
    
    def sketch_midpoint(self, *args) -> Point:
        a, b = args
        return (a + b) * 0.5
    
    def sketch_mirror(self, *args) -> Point:
        a, b = args
        return b * 2 - a
    
    def sketch_nsquare(self, *args) -> Point:
        a, b = args
        ang = -np.pi / 2
        return a + (b - a).rotate(np.sin(ang), np.cos(ang))
    
    def sketch_on_aline(self, *args) -> Line:
        e, d, c, b, a = args
        ab = a - b
        cb = c - b
        de = d - e

        dab = a.distance(b)
        ang_ab = np.arctan2(ab.y / dab, ab.x / dab)

        dcb = c.distance(b)
        ang_bc = np.arctan2(cb.y / dcb, cb.x / dcb)

        dde = d.distance(e)
        ang_de = np.arctan2(de.y / dde, de.x / dde)

        ang_ex = ang_de + ang_bc - ang_ab
        x = e + Point(np.cos(ang_ex), np.sin(ang_ex))
        return Ray(e, x)
    
    def sketch_on_bline(self, *args) -> Line:
        a, b = args
        m = (a + b) * 0.5
        return m.perpendicular_line(Line(a, b))
    
    def sketch_on_circle(self, *args) -> Circle:
        o, a = args
        return Circle(o, o.distance(a))
    
    def sketch_on_line(self, *args) -> Line:
        a, b = args
        return Line(a, b)
        
    def sketch_on_pline(self, *args) -> Line:
        a, b, c = args
        return a.parallel_line(Line(b, c))
    
    def sketch_on_tline(self, *args) -> Line:
        a, b, c = args
        return a.perpendicular_line(Line(b, c))
    
    def sketch_orthocenter(self, *args) -> Point:
        a, b, c = args
        l1 = self.sketch_on_tline(a, b, c)
        l2 = self.sketch_on_tline(b, c, a)
        return line_line_intersection(l1, l2)
    
    def sketch_parallelogram(self, *args) -> Point:
        a, b, c = args
        l1 = self.sketch_on_pline(a, b, c)
        l2 = self.sketch_on_pline(c, a, b)
        return line_line_intersection(l1, l2)
    
    def sketch_pentagon(self, *args) -> list[Point]:
        points = [Point(1.0, 0.0)]
        ang = 0.0

        for i in range(4):
            ang += (2 * np.pi - ang) / (5 - i) * unif(0.5, 1.5)
            point = Point(np.cos(ang), np.sin(ang))
            points.append(point)

        a, b, c, d, e = points  # pylint: disable=unbalanced-tuple-unpacking
        a, b, c, d, e = random_rfss(a, b, c, d, e)
        return [a, b, c, d, e]
    
    def sketch_psquare(self, *args) -> Point:
        a, b = args
        ang = np.pi / 2
        return a + (b - a).rotate(np.sin(ang), np.cos(ang))
    
    def sketch_quadrangle(self, *args) -> list[Point]:
        a = Point(0.0, 0.0)
        b = Point(1.0, 0.0)

        length = np.random.uniform(0.5, 2.0)
        ang = np.random.uniform(np.pi / 3, np.pi * 2 / 3)
        d = head_from(a, ang, length)

        ang = ang_of(b, d)
        ang = np.random.uniform(ang / 10, ang / 9)
        c = head_from(b, ang, length)
        a, b, c, d = random_rfss(a, b, c, d)
        return [a, b, c, d]
    
    def sketch_r_trapezoid(self, *args) -> list[Point]:
        """Right trapezoid with AB horizontal and AD vertical."""
        a = Point(0.0, 1.0)
        d = Point(0.0, 0.0)
        b = Point(unif(0.5, 1.5), 1.0)
        c = Point(unif(0.5, 1.5), 0.0)
        a, b, c, d = random_rfss(a, b, c, d)
        return [a, b, c, d]
    
    def sketch_r_triangle(self, *args) -> list[Point]:
        """Random right triangle with legs on axes."""
        a = Point(0.0, 0.0)
        b = Point(0.0, unif(0.5, 2.0))
        c = Point(unif(0.5, 2.0), 0.0)
        a, b, c = random_rfss(a, b, c)
        return [a, b, c]
    
    def sketch_rectangle(self, *args) -> list[Point]:
        """Axis-aligned rectangle with random width/height."""
        a = Point(0.0, 0.0)
        b = Point(0.0, 1.0)
        l = unif(0.5, 2.0)
        c = Point(l, 1.0)
        d = Point(l, 0.0)
        a, b, c, d = random_rfss(a, b, c, d)
        return [a, b, c, d]
    
    def sketch_reflect(self, *args) -> Point:
        """Reflect point A across line BC."""
        a, b, c = args
        m = a.foot(Line(b, c))
        return m * 2 - a
    
    def sketch_risos(self, *args) -> list[Point]:
        """Right isosceles triangle."""
        a = Point(0.0, 0.0)
        b = Point(0.0, 1.0)
        c = Point(1.0, 0.0)
        a, b, c = random_rfss(a, b, c)
        return [a, b, c]
    
    def sketch_s_angle(self, *args) -> Ray:
        """Ray at point B making angle alpha with BA."""
        a, b, alpha = args
        ang = alpha / 180 * np.pi
        x = b + (a - b).rotatea(ang)
        return Ray(b, x)
    
    def sketch_segment(self, *args) -> list[Point]:
        """Random segment endpoints in [-1,1] box."""
        a = Point(unif(-1, 1), unif(-1, 1))
        b = Point(unif(-1, 1), unif(-1, 1))
        return [a, b]
    
    def sketch_shift(self, *args) -> Point:
        """Translate C by vector BA."""
        c, b, a = args
        return c + (b - a)
    
    def sketch_square(self, *args) -> list[Point]:
        """Square constructed on segment AB."""
        a, b = args
        c = b + (a - b).rotatea(-np.pi / 2)
        d = a + (b - a).rotatea(np.pi / 2)
        return [c, d]
    
    def sketch_isquare(self, *args) -> list[Point]:
        """Axis-aligned unit square, randomly re-ordered."""
        a = Point(0.0, 0.0)
        b = Point(1.0, 0.0)
        c = Point(1.0, 1.0)
        d = Point(0.0, 1.0)
        a, b, c, d = random_rfss(a, b, c, d)
        return [a, b, c, d]
    
    def sketch_trapezoid(self, *args) -> list[Point]:
        """Random trapezoid with AB // CD."""
        d = Point(0.0, 0.0)
        c = Point(1.0, 0.0)

        base = unif(0.5, 2.0)
        height = unif(0.5, 2.0)
        a = Point(unif(0.2, 0.5), height)
        b = Point(a.x + base, height)
        a, b, c, d = random_rfss(a, b, c, d)
        return [a, b, c, d]
    
    def sketch_triangle(self, *args) -> list[Point]:
        """Random triangle."""
        a = Point(0.0, 0.0)
        b = Point(1.0, 0.0)
        ac = unif(0.5, 2.0)
        ang = unif(0.2, 0.8) * np.pi
        c = head_from(a, ang, ac)
        return [a, b, c]
    
    def sketch_triangle12(self, *args) -> list[Point]:
        """Triangle with side-length ratios near 1:2."""
        b = Point(0.0, 0.0)
        c = Point(unif(1.5, 2.5), 0.0)
        a, _ = circle_circle_intersection(Circle(b, 1.0), Circle(c, 2.0))
        a, b, c = random_rfss(a, b, c)
        return [a, b, c]
    
    def sketch_2l1c(self, *args) -> list[Point]:
        """Intersections of perpendiculars from P to AC/BC with circle centered at P."""
        a, b, c, p = args
        bc, ac = Line(b, c), Line(a, c)
        circle = Circle(p, p.distance(a))

        d, d_ = line_circle_intersection(p.perpendicular_line(bc), circle)
        if bc.diff_side(d_, a):
            d = d_

        e, e_ = line_circle_intersection(p.perpendicular_line(ac), circle)
        if ac.diff_side(e_, b):
            e = e_

        df = d.perpendicular_line(Line(p, d))
        ef = e.perpendicular_line(Line(p, e))
        f = line_line_intersection(df, ef)

        g, g_ = line_circle_intersection(Line(c, f), circle)
        if bc.same_side(g_, a):
            g = g_

        b_ = c + (b - c) / b.distance(c)
        a_ = c + (a - c) / a.distance(c)
        m = (a_ + b_) * 0.5
        x = line_line_intersection(Line(c, m), Line(p, g))
        return [x.foot(ac), x.foot(bc), g, x]
    
    def sketch_e5128(self, *args) -> list[Point]:
        """Problem-specific construction e5128."""
        a, b, c, d = args
        g = (a + b) * 0.5
        de = Line(d, g)

        e, f = line_circle_intersection(de, Circle(c, c.distance(b)))

        if e.distance(d) < f.distance(d):
            e = f
        return [e, g]
    
    def sketch_3peq(self, *args) -> list[Point]:
        """Three-point equidistance construction."""
        a, b, c = args
        ab, bc, ca = Line(a, b), Line(b, c), Line(c, a)

        z = b + (c - b) * np.random.uniform(-0.5, 1.5)

        z_ = z * 2 - c
        l = z_.parallel_line(ca)
        x = line_line_intersection(l, ab)
        y = z * 2 - x
        return [x, y, z]
    
    def sketch_trisect(self, *args) -> list[Point]:
        """Trisect angle ABC."""
        a, b, c = args
        ang1 = ang_of(b, a)
        ang2 = ang_of(b, c)

        swap = 0
        if ang1 > ang2:
            ang1, ang2 = ang2, ang1
            swap += 1

        if ang2 - ang1 > np.pi:
            ang1, ang2 = ang2, ang1 + 2 * np.pi
            swap += 1

        angx = ang1 + (ang2 - ang1) / 3
        angy = ang2 - (ang2 - ang1) / 3

        x = b + Point(np.cos(angx), np.sin(angx))
        y = b + Point(np.cos(angy), np.sin(angy))

        ac = Line(a, c)
        x = line_line_intersection(Line(b, x), ac)
        y = line_line_intersection(Line(b, y), ac)

        if swap == 1:
            return [y, x]
        return [x, y]
    
    def sketch_trisegment(self, *args) -> list[Point]:
        """Trisect segment AB."""
        a, b = args
        x, y = a + (b - a) * (1.0 / 3), a + (b - a) * (2.0 / 3)
        return [x, y]
    
    def sketch_on_dia(self, *args) -> Circle:
        """Circle with diameter AB."""
        a, b = args
        o = (a + b) * 0.5
        return Circle(o, o.distance(a))
    
    def sketch_ieq_triangle(self, *args) -> list[Point]:
        a = Point(0.0, 0.0)
        b = Point(1.0, 0.0)

        c, _ = Circle(a, a.distance(b)).intersect(Circle(b, b.distance(a)))
        return [a, b, c]
    
    def sketch_on_opline(self, *args) -> Ray:
        a, b = args
        return Ray(a, a + a - b)
    
    def sketch_cc_tangent(self, *args) -> list[Point]:
        o, a, w, b = args
        ra, rb = o.distance(a), w.distance(b)

        ow = Line(o, w)
        if close_enough(ra, rb):
            oo = ow.perpendicular_line(o)
            oa = Circle(o, ra)
            x, z = line_circle_intersection(oo, oa)
            y = x + w - o
            t = z + w - o
            return [x, y, z, t]
    
    def sketch_cc_tangent0(self, *args) -> Ray:
        o, a, w, b = args
        return self.sketch_cc_tangent(o, a, w, b)[:2]
    
    def sketch_eqangle3(self, *args) -> list[Point]:
        a, b, d, e, f = args
        de = d.distance(e)
        ef = e.distance(f)
        ab = b.distance(a)
        ang_ax = ang_of(a, b) + ang_between(e, d, f)
        x = head_from(a, ang_ax, length=de / ef * ab)   
        o = self.sketch_circle(a, b, x)
        return Circle(o, o.distance(a))
    
    def sketch_tangent(self, *args) -> list[Point]:
        a, o, b = args
        dia = self.sketch_dia([a, o])
        return list(circle_circle_intersection(Circle(o, o.distance(b)), dia))
    
    def sketch_on_circum(self, *args) -> Circle:
        a, b, c = args
        o = self.sketch_circle(a, b, c)
        return Circle(o, o.distance(a))
    
    def sketch_sameside(self, *args) -> HalfPlane:
        a, b, c = args
        return HalfPlane(a, b, c)
    
    def sketch_opposingsides(self, *args) -> HalfPlane:
        a, b, c = args
        return HalfPlane(a, b, c, opposingsides=True)
    
    def reduce(self, objs, existing_points) -> list[Point]:
        """Reduce intersecting objects into sampled intersection points.

        Filters half-planes, handles point-only cases, samples within half-planes,
        or intersects pairs of essential geometric objects.
        """
        essential_objs = [i for i in objs if not isinstance(i, HalfPlane)]
        halfplane_objs = [i for i in objs if isinstance(i, HalfPlane)]
  
        if all(isinstance(o, Point) for o in objs):
            return objs
  
        elif all(isinstance(o, HalfPlane) for o in objs):
            if len(objs) == 1:
                return objs[0].sample_within_halfplanes(existing_points,[])
            else:
                return objs[0].sample_within_halfplanes(existing_points,objs[1:])

        elif len(essential_objs) == 1:
            if not halfplane_objs:
                return objs[0].sample_within(existing_points)
            else:
                return objs[0].sample_within_halfplanes(existing_points,halfplane_objs)
  
        elif len(essential_objs) == 2:
            a, b = essential_objs
            result = a.intersect(b)
            
            if isinstance(result, Point):
                if halfplane_objs and not all(i.contains(result) for i in halfplane_objs):
                    raise Exception()
                return [result]
            
            a, b = result
            
            if halfplane_objs:
                a_correct_side = all(i.contains(a) for i in halfplane_objs)
                b_correct_side = all(i.contains(b) for i in halfplane_objs)
                
                if a_correct_side and not b_correct_side:
                    return [a]
                elif b_correct_side and not a_correct_side:
                    return [b]
                elif not a_correct_side and not b_correct_side:
                    raise Exception()
                        
            a_close = any([a.close(x) for x in existing_points])
            b_close = any([b.close(x) for x in existing_points])
            
            if a_close and not b_close:
                return [b]
            
            elif b_close and not a_close:
                return [a]
            else:
                return [np.random.choice([a, b])]
    
    def draw(self, new_points, construction):
        func = getattr(self, 'draw_' + construction.__class__.__name__[10:])
        args = [arg if isinstance(arg, float) else self.name2point[arg.name] for arg in construction.arguments()]
        func(*new_points, *args)
    
    def draw_angle_bisector(self, *args):
        x, a, b, c = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(b, x))
    
    def draw_angle_mirror(self, *args):
        x, a, b, c = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(b, x))
    
    def draw_circle(self, *args):
        x, a, b, c = args
        # self.segments.append(Segment(a, x))
        # self.segments.append(Segment(b, x))
        # self.segments.append(Segment(c, x))
        self.circles.append(Circle(x, x.distance(a)))
        
    def draw_circumcenter(self, *args):
        x, a, b, c = args
        # self.segments.append(Segment(a, x))
        # self.segments.append(Segment(b, x))
        # self.segments.append(Segment(c, x))
        self.circles.append(Circle(x, x.distance(a)))
        
    def draw_eq_quadrangle(self, *args):
        a, b, c, d = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, d))
        self.segments.append(Segment(d, a))
        
    def draw_eq_trapezoid(self, *args):
        a, b, c, d = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, d))
        self.segments.append(Segment(d, a))
        
    def draw_eq_triangle(self, *args):
        x, b, c = args
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, x))
        self.segments.append(Segment(x, b))
    
    def draw_eqangle2(self, *args):
        x, a, b, c = args
        self.segments.append(Segment(b, a))
        self.segments.append(Segment(a, x))
        self.segments.append(Segment(c, b))
        self.segments.append(Segment(b, b))
    
    def draw_eqdia_quadrangle(self, *args):
        a, b, c, d = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, d))
        self.segments.append(Segment(d, a))
        self.segments.append(Segment(b, d))
        self.segments.append(Segment(a, c))
        
    def draw_eqdistance(self, *args):
        x, a, b, c = args
        self.segments.append(Segment(x, a))
        self.segments.append(Segment(b, c))
    
    def draw_eqdistance2(self, *args):
        x, a, b, c, alpha = args
        self.segments.append(Segment(x, a))
        self.segments.append(Segment(b, c))
        
    def draw_eqdistance2(self, *args):
        x, a, alpha = args
        self.segments.append(Segment(x, a))
    
    def draw_foot(self, *args):
        x, a, b, c = args
        self.segments.append(Segment(x, a))
        self.segments.append(Segment(x, b))
        self.segments.append(Segment(x, c))
        self.segments.append(Segment(b, c))
    
    def draw_free(self, *args):
        x = args
        
    def draw_incenter(self, *args):
        i, a, b, c = args
        x = i.foot(Line(b, c))
        y = i.foot(Line(c, a))
        z = i.foot(Line(a, b))
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, a))
        self.circles.append(Circle(p1=x, p2=y, p3=z))
        
    def draw_incenter2(self, *args):
        x, y, z, i, a, b, c = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, a))
        self.circles.append(Circle(p1=x, p2=y, p3=z))
    
    def draw_excenter(self, *args):
        i, a, b, c = args
        x = i.foot(Line(b, c))
        y = i.foot(Line(c, a))
        z = i.foot(Line(a, b))
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, a))
        self.circles.append(Circle(p1=x, p2=y, p3=z))
        
    def draw_excenter2(self, *args):
        x, y, z, i, a, b, c = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, a))
        self.circles.append(Circle(p1=x, p2=y, p3=z))
        
    def draw_centroid(self, *args):
        x, y, z, i, a, b, c = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, a))
        self.segments.append(Segment(a, x))
        self.segments.append(Segment(b, y))
        self.segments.append(Segment(c, z))
    
    def draw_intersection_cc(self, *args):
        x, o, w, a = args
        self.segments.append(Segment(o, a))
        self.segments.append(Segment(o, x))
        self.segments.append(Segment(w, a))
        self.segments.append(Segment(w, x))
        self.circles.append(Circle(o, o.distance(a)))
        self.circles.append(Circle(w, w.distance(a)))
    
    def draw_intersection_lc(self, *args):
        x, a, o, b = args
        self.segments.append(Segment(x, a))
        self.segments.append(Segment(x, b))
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(o, b))
        self.segments.append(Segment(o, x))
        self.circles.append(Circle(o, o.distance(b)))
        
    def draw_intersection_ll(self, *args):
        x, a, b, c, d = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(c, d))
        self.segments.append(Segment(a, x))
        self.segments.append(Segment(b, x))
        self.segments.append(Segment(c, x))
        self.segments.append(Segment(d, x))
        
    def draw_intersection_lp(self, *args):
        x, a, b, c, m, n = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(m, n))
        self.segments.append(Segment(a, x))
        self.segments.append(Segment(b, x))
        self.segments.append(Segment(c, x))
        
    def draw_intersection_lp(self, *args):
        x, a, b, c, m, n = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(m, n))
        self.segments.append(Segment(a, x))
        self.segments.append(Segment(b, x))
        self.segments.append(Segment(c, x))
        
    def draw_intersection_lt(self, *args):
        x, a, b, c, d, e = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(d, e))
        self.segments.append(Segment(a, x))
        self.segments.append(Segment(b, x))
        self.segments.append(Segment(c, x))
        
    def draw_intersection_pp(self, *args):
        x, a, b, c, d, e, f = args
        self.segments.append(Segment(a, x))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(d, x))
        self.segments.append(Segment(e, f))
    
    def draw_intersection_tt(self, *args):
        x, a, b, c, d, e, f = args
        self.segments.append(Segment(a, x))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(d, x))
        self.segments.append(Segment(e, f))
    
    def draw_iso_triangle(self, *args):
        a, b, c = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, a))
        
    def draw_lc_tangent(self, *args):
        x, a, o = args
        self.segments.append(Segment(a, x))
        self.segments.append(Segment(a, o))
        self.circles.append(Circle(o, o.distance(a)))
    
    def draw_midpoint(self, *args):
        x, a, b = args
        self.segments.append(Segment(a, x))
        self.segments.append(Segment(b, x))
    
    def draw_mirror(self, *args):
        x, a, b = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, x))
        
    def draw_nsquare(self, *args):
        x, a, b = args
        self.segments.append(Segment(x, a))
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, x))
    
    def draw_on_aline(self, *args):
        x, a, b, c, d, e = args
        self.segments.append(Segment(e, d))
        self.segments.append(Segment(d, c))
        self.segments.append(Segment(b, a))
        self.segments.append(Segment(a, x))
    
    def draw_on_bline(self, *args):
        x, a, b = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(a, x))
        self.segments.append(Segment(b, x))
    
    def draw_on_circle(self, *args):
        x, o, a = args
        self.segments.append(Segment(o, a))
        self.segments.append(Segment(o, x))
        self.circles.append(Circle(o, o.distance(x)))
        
    def draw_on_line(self, *args):
        x, a, b = args
        self.segments.append(Segment(x, a))
        self.segments.append(Segment(x, b))
        self.segments.append(Segment(a, b))
        
    def draw_on_pline(self, *args):
        x, a, b, c = args
        self.segments.append(Segment(x, a))
        self.segments.append(Segment(b, c))
    
    def draw_on_tline(self, *args):
        x, a, b, c = args
        self.segments.append(Segment(x, a))
        self.segments.append(Segment(b, c))
    
    def draw_orthocenter(self, *args):
        x, a, b, c = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, a))
    
    def draw_parallelogram(self, *args):
        x, a, b, c = args
        self.segments.append(Segment(x, a))
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, x))
    
    def draw_parallelogram(self, *args):
        x, a, b, c = args
        self.segments.append(Segment(x, a))
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, x))
    
    def draw_pentagon(self, *args):
        a, b, c, d, e = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, d))
        self.segments.append(Segment(d, e))
        self.segments.append(Segment(e, a))
    
    def draw_psquare(self, *args):
        x, a, b = args
        self.segments.append(Segment(x, a))
        self.segments.append(Segment(a, b))
        
    def draw_quadrangle(self, *args):
        a, b, c, d = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, d))
        self.segments.append(Segment(d, a))
    
    def draw_r_trapezoid(self, *args):
        a, b, c, d = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, d))
        self.segments.append(Segment(d, a))
    
    def draw_r_triangle(self, *args):
        a, b, c = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, a))
        
    def draw_rectangle(self, *args):
        a, b, c, d = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, d))
        self.segments.append(Segment(d, a))
    
    def draw_reflect(self, *args):
        x, a, b, c = args
        self.segments.append(Segment(b, c))
    
    def draw_risos(self, *args):
        a, b, c = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, a))
    
    def draw_s_angle(self, *args):
        x, a, b, alpha = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, x))
    
    def draw_segment(self, *args):
        a, b = args
        self.segments.append(Segment(a, b))
    
    def draw_s_segment(self, *args):
        a, b, alpha = args
        self.segments.append(Segment(a, b))
        
    def draw_shift(self, *args):
        x, b, c, d = args
        self.segments.append(Segment(x, b))
        self.segments.append(Segment(c, d))
        self.segments.append(Segment(x, c))
        self.segments.append(Segment(b, d))
    
    def draw_square(self, *args):
        x, y, a, b = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, x))
        self.segments.append(Segment(x, y))
        self.segments.append(Segment(y, a))
    
    def draw_isquare(self, *args):
        a, b, c, d = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, d))
        self.segments.append(Segment(d, a))
    
    def draw_trapezoid(self, *args):
        a, b, c, d = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, d))
        self.segments.append(Segment(d, a))
    
    def draw_triangle(self, *args):
        a, b, c = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, a))
    
    def draw_triangle12(self, *args):
        a, b, c = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, a))
    
    def draw_2l1c(self, *args):
        x, y, z, i, a, b, c, o = args
        self.segments.append(Segment(a, c))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(a, o))
        self.segments.append(Segment(b, o))
        
        self.segments.append(Segment(i, x))
        self.segments.append(Segment(i, y))
        self.segments.append(Segment(i, z))
        
        self.segments.append(Segment(c, x))
        self.segments.append(Segment(a, x))
        self.segments.append(Segment(c, y))
        self.segments.append(Segment(b, y))
        self.segments.append(Segment(o, z))
        
        self.circles.append(Circle(i, i.distance(x)))
        self.circles.append(Circle(o, o.distance(a)))
    
    def draw_e5128(self, *args):
        x, y, a, b, c, d = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, d))
        self.segments.append(Segment(d, a))
        
        self.segments.append(Segment(a, x))
        self.segments.append(Segment(x, y))
        self.segments.append(Segment(c, x))
    
    def draw_3peq(self, *args):
        x, y, z, a, b, c = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, a))
        
        self.segments.append(Segment(a, x))
        self.segments.append(Segment(b, x))
        
        self.segments.append(Segment(a, y))
        self.segments.append(Segment(c, y))
        
        self.segments.append(Segment(c, z))
        self.segments.append(Segment(b, z))
        
        self.segments.append(Segment(x, y))
        self.segments.append(Segment(y, z))
        self.segments.append(Segment(z, x))
    
    def draw_trisect(self, *args):
        x, y, a, b, c = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        
        self.segments.append(Segment(b, x))
        self.segments.append(Segment(b, y))
        
        self.segments.append(Segment(a, x))
        self.segments.append(Segment(x, y))
        self.segments.append(Segment(y, c))
        
    def draw_trisegment(self, *args):
        x, y, a, b = args
        self.segments.append(Segment(a, x))
        self.segments.append(Segment(x, y))
        self.segments.append(Segment(y, b))
    
    def draw_on_dia(self, *args):
        x, a, b = args
        self.segments.append(Segment(x, a))
        self.segments.append(Segment(x, b))
        
    def draw_ieq_triangle(self, *args):
        a, b, c = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(b, c))
        self.segments.append(Segment(c, a))
    
    def draw_on_opline(self, *args):
        x, a, b = args
        self.segments.append(Segment(a, b))
        self.segments.append(Segment(x, a))
        self.segments.append(Segment(x, b))
    
    def draw_cc_tangent0(self, *args):
        x, y, o, a, w, b = args
        self.segments.append(Segment(o, a))
        self.segments.append(Segment(o, x))
        
        self.segments.append(Segment(w, b))
        self.segments.append(Segment(w, y))
        
        self.segments.append(Segment(x, y))
        
        self.circles.append(Circle(o, o.distance(a)))
        self.circles.append(Circle(w, w.distance(b)))
    
    def draw_cc_tangent(self, *args):
        x, y, z, i, o, a, w, b = args
        self.segments.append(Segment(o, a))
        self.segments.append(Segment(o, x))
        self.segments.append(Segment(o, z))
        
        self.segments.append(Segment(w, b))
        self.segments.append(Segment(w, y))
        self.segments.append(Segment(w, i))
        
        self.segments.append(Segment(x, y))
        self.segments.append(Segment(z, i))
        
        self.circles.append(Circle(o, o.distance(a)))
        self.circles.append(Circle(w, w.distance(b)))
    
    def draw_eqangle3(self, *args):
        x, a, b, d, e, f = args
        self.segments.append(Segment(f, d))
        self.segments.append(Segment(d, e))
        
        self.segments.append(Segment(a, x))
        self.segments.append(Segment(x, b))
    
    def draw_tangent(self, *args):
        x, y, a, o, b = args
        self.segments.append(Segment(o, b))
        self.segments.append(Segment(o, x))
        self.segments.append(Segment(o, y))
        self.segments.append(Segment(a, x))
        self.segments.append(Segment(a, y))
        
        self.circles.append(Circle(o, o.distance(b)))
    
    def draw_on_circum(self, *args):
        x, a, b, c = args
        self.circles.append(Circle(p1=a, p2=b, p3=c))
    
    def draw_sameside(self, *args):
        x, a, b, c = args
    
    def draw_opposingsides(self, *args):
        x, a, b, c = args
        
    def draw_diagram(self, show=False):
        """Draw the current diagram; optionally display the matplotlib figure."""
        imsize = 512 / 100
        self.fig, self.ax = plt.subplots(figsize=(imsize, imsize), dpi=300)
        self.ax.set_facecolor((1.0, 1.0, 1.0))
        
        for segment in self.segments:
            p1, p2 = segment.p1, segment.p2
            lx, ly = (p1.x, p2.x), (p1.y, p2.y)
            self.ax.plot(lx, ly, color='black', lw=1.2, alpha=0.8, ls='-')
        
        for circle in self.circles:
            self.ax.add_patch(
                plt.Circle(
                    (circle.center.x, circle.center.y),
                    circle.radius,
                    color='red',
                    alpha=0.8,
                    fill=False,
                    lw=1.2,
                    ls='-'
                )
            )
        
        for p in self.points:
            self.ax.scatter(p.x, p.y, color='black', s=15)
            self.ax.annotate(self.point2name[p], (p.x+0.015, p.y+0.015), color='black', fontsize=8)
        
        self.ax.set_aspect('equal')
        self.ax.set_axis_off()
        self.fig.subplots_adjust(left=0.05, right=0.95, top=0.95, bottom=0.05, wspace=0, hspace=0)
        xmin = min([p.x for p in self.points])
        xmax = max([p.x for p in self.points])
        ymin = min([p.y for p in self.points])
        ymax = max([p.y for p in self.points])
        x_margin = (xmax - xmin) * 0.1
        y_margin = (ymax - ymin) * 0.1
        
        self.ax.margins(x_margin, y_margin)
        
        self.save_diagram()
        
        if show:
            plt.show()
            
        plt.close(self.fig)
    
    def save_diagram(self):
        if self.save_path is not None:
            parent_dir = os.path.dirname(self.save_path)
            if parent_dir and not os.path.exists(parent_dir):
                os.makedirs(parent_dir)
            self.fig.savefig(self.save_path)
        


"""
Legacy shim so existing workflows can still run `pip install .`.
All metadata now lives in pyproject.toml and is handled by hatchling.
"""
from setuptools import setup

if __name__ == "__main__":
    setup()


"""Deductive database for geometric inference rules.

Matches registered inference rules against the current `State`, instantiates
them with concrete points, and applies resulting conclusions. Uses a Z3 encoding
to search admissible assignments while preserving canonical point orderings via
`Lt` constraints.
"""

import z3
import math
            
from z3 import Solver, BitVec, BitVecVal, ULT, And, Or
from tqdm import tqdm

from pyeuclid.formalization.state import *
from pyeuclid.engine.inference_rule import *


class DeductiveDatabase:
    """Match and apply geometric inference rules against a state."""

    def __init__(self, state, inner_theorems=inference_rule_sets["ex"], outer_theorems=inference_rule_sets["basic"]):
        """Initialize a deductive database.

        Args:
            state (State): Shared state being reasoned over.
            inner_theorems (Iterable[type[InferenceRule]]): Higher-priority rules applied exhaustively first.
            outer_theorems (Iterable[type[InferenceRule]]): Secondary rules applied after inner closure.
        """
        self.state = state
        self.inner_theorems = inner_theorems
        self.outer_theorems = outer_theorems
        self.closure = False
        
    def get_applicable_theorems(self, theorems):
        """Instantiate and return all applicable rules from the given set.

        Returns:
            list[InferenceRule]: Concrete rule instances whose conditions are satisfiable in the current state.
        """
        def search_assignments(theorem):
            if not theorem in self.state.solvers:
                self.state.solvers[theorem] = Solver()
            solver = self.state.solvers[theorem]
            solver.push()
            slots = theorem.__init__.__annotations__
            formal_entities = {}
            for key, attr_type in slots.items():
                if isinstance(attr_type, str):
                    assert attr_type == 'Point'
                else:
                    assert attr_type.__name__ == "Point"
                formal_entities[key] = Point(key)
            example = theorem(**formal_entities)
            formal_conditions = example.condition()
            
            nbits = math.ceil(math.log2(len(self.state.points)))
            point_encoding = {}
            point_decoding = {}
            formal_points = {}
            points = list(self.state.points)
            points.sort(key=lambda x: x.name)
            for i, point in enumerate(points):
                point_encoding[point.name] = BitVecVal(i, nbits)
                point_decoding[BitVecVal(i, nbits)] = point
            for name in formal_entities:
                formal_points[name] = BitVec(name, nbits)
                if len(self.state.points) < 2**nbits:
                    solver.add(ULT(formal_points[name], len(self.state.points)))

            def in_component(formal, component):
                clause = False
                if len(formal) == 2:
                    for item in component:
                        actual, _ = get_points_and_symbols(item)
                        p1 = And(formal_points[formal[0]] == point_encoding[actual[0]],
                                formal_points[formal[1]] == point_encoding[actual[1]])
                        p2 = And(formal_points[formal[0]] == point_encoding[actual[1]],
                                formal_points[formal[1]] == point_encoding[actual[0]])
                        clause = Or(clause, Or(p1, p2))
                elif len(formal) == 3:
                    for item in component:
                        actual, _ = get_points_and_symbols(item)
                        p1 = And(formal_points[formal[0]] == point_encoding[actual[0]], formal_points[formal[1]]
                                == point_encoding[actual[1]], formal_points[formal[2]] == point_encoding[actual[2]])
                        p2 = And(formal_points[formal[0]] == point_encoding[actual[2]], formal_points[formal[1]]
                                == point_encoding[actual[1]], formal_points[formal[2]] == point_encoding[actual[0]])
                        clause = Or(clause, Or(p1, p2))
                else:
                    assert False
                return clause

            for cond in formal_conditions:
                clause = False
                if isinstance(cond, Relation):
                    formal = cond.get_points()
                else:
                    formal, _ = get_points_and_symbols(cond)
                if isinstance(cond, Relation):
                    if isinstance(cond, Equal):
                        clause = formal_points[cond.v1.name] == formal_points[cond.v2.name]
                        if cond.negated:
                            clause = z3.Not(clause)
                    elif isinstance(cond, Lt):
                        clause = ULT(
                            formal_points[cond.v1.name], formal_points[cond.v2.name])
                    else:
                        assert type(cond) in (
                            Collinear, SameSide, Between, Perpendicular, Concyclic, Parallel)
                        if type(cond) == Between:
                            clauses = []
                            for rel in self.state.relations:
                                if type(rel) == type(cond):
                                    assert not rel.negated
                                    permutations = rel.permutations()
                                    for perm in permutations:
                                        assignment = True
                                        for i in range(len(formal)):
                                            assignment = And(
                                                formal_points[formal[i]] == point_encoding[perm[i]], assignment)
                                        clauses.append(assignment)
                            clause = Or(*clauses)
                            if cond.negated:  # we have all between relations, and never store negated between relations
                                clause = z3.Not(clause)
                        else:
                            for rel in self.state.relations:
                                if type(rel) == type(cond) and rel.negated == cond.negated:
                                    if hasattr(rel, "permutations"):
                                        permutations = rel.permutations()
                                    else:
                                        permutations = [
                                            re.pattern.findall(str(rel))]
                                    for perm in permutations:
                                        if not isinstance(perm[0], str):
                                            perm = [item.name for item in perm]
                                        partial_assignment = True
                                        for i in range(len(formal)):
                                            partial_assignment = And(
                                                formal_points[formal[i]] == point_encoding[perm[i]], partial_assignment)
                                        clause = Or(partial_assignment, clause)
                            if type(cond) == Collinear:
                                degenerate = Or(formal_points[formal[0]]==formal_points[formal[1]], formal_points[formal[1]]==formal_points[formal[2]], formal_points[formal[2]]==formal_points[formal[0]])
                                clause = Or(clause, degenerate)
                elif isinstance(cond, sympy.core.expr.Expr):
                    pattern_eqlength = re.compile(r"^-?Length\w+ [-\+] Length\w+$")
                    pattern_eqangle = re.compile(r"^-?Angle\w+ [-\+] Angle\w+$")
                    pattern_eqratio = re.compile(
                        r"^-?Length\w+/Length\w+ [\+-] Length\w+/Length\w+$")
                    pattern_angle_const = re.compile(
                        r"^-?Angle\w+ [-\+] [\w/\d]+$")
                    pattern_angle_sum = re.compile(
                        r"^-?Angle\w+ [-\+] Angle\w+ [-\+] [\w/\d]+$")
                    s = str(cond)
                    if pattern_eqlength.match(s):
                        points, _ = get_points_and_symbols(cond)
                        l, r = points[:2], points[2:]
                        for component in self.state.lengths.equivalence_classes().values():
                            clause = Or(clause, And(in_component(
                                l, component), in_component(r, component)))
                    elif pattern_eqangle.match(s):
                        points, _ = get_points_and_symbols(cond)
                        l, r = points[:3], points[3:]
                        for component in self.state.angles.equivalence_classes().values():
                            clause = Or(clause, And(in_component(
                                l, component), in_component(r, component)))
                    elif pattern_eqratio.match(s):
                        points, _ = get_points_and_symbols(cond)
                        a, b, c, d = points[:2], points[2:4], points[4:6], points[6:8]
                        for ratios in self.state.ratios.values():
                            l_clause = False
                            r_clause = False
                            for ratio in ratios:
                                _, symbols = get_points_and_symbols(ratio)
                                length1, length2 = symbols
                                length1, length2 = self.state.lengths.find(
                                    length1), self.state.lengths.find(length2)
                                component1, component2 = self.state.lengths.equivalence_classes(
                                )[length1], self.state.lengths.equivalence_classes()[length2]
                                l_clause = Or(l_clause, And(in_component(
                                    a, component1), in_component(b, component2)))
                                r_clause = Or(r_clause, And(in_component(
                                    c, component1), in_component(d, component2)))
                            clause = Or(clause, And(l_clause, r_clause))
                    elif pattern_angle_const.match(s):
                        points, _ = get_points_and_symbols(cond)
                        left = points[:3]
                        cnst = [arg for arg in cond.args if len(arg.free_symbols)==0][0]
                        cnst = abs(cnst)
                        for rep, component in self.state.angles.equivalence_classes().items():
                            if self.state.check_conditions(cnst - rep):
                                clause = in_component(left, component)
                                break
                    else:
                        assert pattern_angle_sum.match(s)
                        cnst = [arg for arg in cond.args if len(
                            arg.free_symbols) == 0][0]
                        cnst = abs(cnst)
                        points, _ = get_points_and_symbols(cond)
                        left, right = points[:3], points[3:]
                        for rep, angle_sums in self.state.angle_sums.items():
                            if self.state.check_conditions(cnst-rep):
                                for angle_sum in angle_sums:
                                    if isinstance(angle_sum, sympy.core.add.Add):
                                        angle1, angle2 = angle_sum.args
                                        # angle1 + angle2
                                    else:
                                        angle1 = list(angle_sum.free_symbols)[0]
                                        angle2 = angle1
                                        # 2 * angle_1
                                    angle1, angle2 = self.state.angles.find(angle1), self.state.angles.find(angle2)
                                    try:
                                        component1, component2 = self.state.angles.equivalence_classes(
                                        )[angle1], self.state.angles.equivalence_classes()[angle2]
                                    except:
                                        breakpoint()
                                        assert False
                                    clause = Or(clause, And(in_component(
                                        left, component1), in_component(right, component2)))
                                break
                solver.add(clause)
            solutions = []
            assignments = []
            while solver.check() == z3.sat:
                m = solver.model()
                dic = {str(i): point_decoding[m[i]] for i in m}
                concrete = theorem(**dic)
                concrete._depth = self.state.current_depth
                # if try complex, solutions in later iterations may be weaker than previous ones and unionfind because of abondoning complex equations, causing check condition failure
                if not self.state.try_complex and not self.state.check_conditions(concrete.condition()):
                    for condition in concrete.condition():
                        if not self.state.check_conditions(condition):
                            print(f"Failed condition: {condition}")
                            breakpoint()
                            self.state.check_conditions(condition)
                            assert False
                if not concrete.degenerate():
                    assignments.append(concrete)
                solution = False
                for i in m:
                    solution = Or((formal_points[str(i)] != m[i]), solution)
                solver.add(solution)
                solutions.append(solution)
            solver.pop()
            for item in solutions:
                solver.add(item)
            solver.push()
            assignments.sort(key=lambda x: str(x))
            return assignments
        
        applicable_theorems = []
        pbar = tqdm(theorems, disable=self.state.silent)
        for theorem in pbar:
            pbar.set_description(
                f"{theorem.__name__} #rels {len(self.state.relations)} # eqns {len(self.state.equations)}")
            concrete_theorems = search_assignments(theorem)
            applicable_theorems += concrete_theorems
        return applicable_theorems
    
    def apply(self, inferences):
        """Apply instantiated rules by adding their conclusions to the state.

        Args:
            inferences (Iterable[InferenceRule]): Instantiated rules to apply.

        Returns:
            None
        """
        last = None
        cnt = 0
        for item in inferences:
            tmp = type(item)
            if not tmp == last:
                if cnt > 3:
                    if not self.state.silent:
                        self.state.logger.info(f"...and {cnt-3} more.")
                cnt = 0
                last = tmp
            if cnt < 3:
                if not self.state.silent:
                    self.state.logger.info(str(item))
            cnt += 1
            conclusions = item.conclusion()
            for i, conclusion in enumerate(conclusions):
                if isinstance(conclusion, sympy.core.expr.Expr):
                    conclusion = Traced(conclusion)
                    conclusion.sources = [item]
                else:
                    conclusion.source = item
                conclusion.depth = self.state.current_depth
                item.depth = self.state.current_depth
                conclusions[i] = conclusion
            self.state.add_relations(conclusions)
        if cnt > 3:
            if not self.state.silent:
                self.state.logger.info(f"...and {cnt - 3} more.")
    
    def run(self):
        """Execute one deductive phase over inner then outer theorems.

        Updates `closure` when no further inferences are available.

        Returns:
            None
        """
        inner_closure = True
        while True:
            if self.state.complete() is not None:
                return
            inner_applicable = self.get_applicable_theorems(self.inner_theorems)
            self.apply(inner_applicable)
            if len(inner_applicable) == 0:
                break
            inner_closure = False
            
        if self.state.complete() is not None:
            return
        
        applicable_theorems = self.get_applicable_theorems(self.outer_theorems)
        self.apply(applicable_theorems)
        
        if len(applicable_theorems) == 0 and inner_closure:
            self.closure = True
            if not self.state.silent:
                self.state.logger.debug("Found Closure")
            return


import unittest
import time

import os
from sympy import sympify

from pyeuclid.formalization.state import State
from pyeuclid.formalization.relation import *
from pyeuclid.formalization.translation import parse_texts_from_file
from pyeuclid.engine.inference_rule import inference_rule_sets
from pyeuclid.engine.deductive_database import DeductiveDatabase
from pyeuclid.engine.algebraic_system import AlgebraicSystem
from pyeuclid.engine.proof_generator import ProofGenerator
from pyeuclid.engine.engine import Engine
import traceback
from stopit import ThreadingTimeout as Timeout

class TestBenchmarks(unittest.TestCase):
    def test_jgex_ag_231(self):
        rank = int(os.environ.get("OMPI_COMM_WORLD_RANK", 0))
        world_size = int(os.environ.get("OMPI_COMM_WORLD_SIZE", 1))
        texts = parse_texts_from_file('data/JGEX-AG-231.txt')
        for idx, text in enumerate(texts):
            if not idx%world_size == rank:
                continue
            state = State()
            if world_size > 1:
                state.silent = True
            try:
                state.load_problem_from_text(text, f'diagrams/JGEX-AG-231/{idx+1}.jpg')
                deductive_database = DeductiveDatabase(state)
                algebraic_system = AlgebraicSystem(state)
                proof_generator = ProofGenerator(state)
                engine = Engine(state, deductive_database, algebraic_system)
                t = time.time()
                with Timeout(3600):
                    engine.search()
                t = time.time() - t
                if state.complete() is not None:
                    print(f"{idx} solved in {t} seconds")
                    proof_generator.generate_proof()
                    if world_size == 1:
                        proof_generator.show_proof()
                else:
                    print(f"{idx} unsolved in {t} seconds")
            except BaseException as e:
                if isinstance(e, KeyboardInterrupt):
                    exit()
                print(f"{idx} error {text} {e}")
                print(traceback.format_exc())
            
    def test_geometry3k(self):
        rank = int(os.environ.get("OMPI_COMM_WORLD_RANK", 0))
        world_size = int(os.environ.get("OMPI_COMM_WORLD_SIZE", 1))
        for idx in range(2401, 3002):
            if not idx%world_size == rank:
                continue
            if not os.path.isfile(f"data/Geometry3K/{idx}/problem.py"):
                continue
            namespace = {}
            try:
                with open(f'data/Geometry3K/{idx}/problem.py', "r") as file:
                    exec(file.read(), namespace)
                conditions = namespace.get("conditions")
                goal = namespace.get("goal")
                solution = namespace.get("solution")
                diagrammatic_relations = namespace.get("new_diagrammatic_relations")
                for i in conditions:
                    if isinstance(i, Between):
                        diagrammatic_relations.discard(NotCollinear(i.p1,i.p2,i.p3))
                state = State()
                state.silent = True
                state.load_problem(conditions=conditions, goal=goal)
                state.add_relations(list(diagrammatic_relations))
                
                deductive_database = DeductiveDatabase(state, outer_theorems=inference_rule_sets['basic']+inference_rule_sets['complex'])
                algebraic_system = AlgebraicSystem(state)
                proof_generator = ProofGenerator(state)
                engine = Engine(state, deductive_database, algebraic_system)

                try:
                    t = time.time()
                    with Timeout(600) as tt:
                        engine.search()
                    result = state.complete()
                    if result is None:
                        state.try_complex = True
                        engine.deductive_database.closure = False
                        with Timeout(600) as tt:
                            engine.search()
                    t = time.time() - t
                except:
                    pass
                result = state.complete()
                
                if result is not None:
                    if (result is True or abs((sympify(result).evalf() - sympify(solution).evalf()) / (sympify(solution).evalf() + 1e-4)) < 2e-2):
                        print(f"{idx} solved in {t} seconds {state.try_complex}")
                    else:
                        print(f"{idx} wrong solution in {t} seconds")
                    proof_generator.generate_proof()
                    if world_size == 1:
                        proof_generator.show_proof()
                else:
                    print(f"{idx} unsolved in {t} seconds")
            except BaseException as e:
                if isinstance(e, KeyboardInterrupt):
                    exit()
                print(f"{idx} error {e}")
                print(traceback.format_exc())

if __name__ == '__main__':
    unittest.main()